# Yuio project, MIT license.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

"""
Querying terminal info and working with ANSI terminals.

This is a low-level module upon which :mod:`yuio.io` builds
its higher-level abstraction.


Detecting terminal capabilities
-------------------------------

Terminal capabilities are stored in a :class:`Term` object.

Usually, you don't need to query terminal capabilities yourself,
as you can use Yuio's global configuration from :mod:`yuio.io`
(see :func:`yuio.io.get_term`).

However, you can get a :class:`Term` object by using :func:`get_term_from_stream`:

.. autofunction:: get_term_from_stream

.. autofunction:: get_tty

.. autoclass:: Term
   :members:

.. autoclass:: TerminalTheme
   :members:

.. autoclass:: Lightness
   :members:


Utilities
---------

.. autofunction:: stream_is_unicode

.. autofunction:: get_tty_size

.. autofunction:: detect_ci

.. autofunction:: detect_ci_color_support


Re-imports
----------

.. type:: ColorSupport
    :no-index:

    Alias of :obj:`yuio.color.ColorSupport`.

"""

from __future__ import annotations

import atexit
import contextlib
import dataclasses
import enum
import io
import locale
import os
import re
import shutil
import sys
import threading
from dataclasses import dataclass

import yuio
import yuio.color
from yuio.color import ColorSupport

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import typing_extensions as _t
else:
    from yuio import _typing as _t

__all__ = [
    "ColorSupport",
    "Lightness",
    "Term",
    "TerminalTheme",
    "detect_ci",
    "detect_ci_color_support",
    "get_term_from_stream",
    "get_tty",
    "get_tty_size",
    "stream_is_unicode",
]

T = _t.TypeVar("T")


_LOCK = threading.Lock()

# These variables are set in `_prepare_tty`.
_TTY_SETUP_PERFORMED: bool = False
_TTY_OUTPUT: _t.TextIO | None
_TTY_INPUT: _t.TextIO | None
_TERMINAL_THEME: TerminalTheme | None
_EXPLICIT_COLOR_SUPPORT: ColorSupport | bool | None
_COLOR_SUPPORT: ColorSupport


# Redefine canonical streams so that we don't monkeypatch `sys.__std*__` in tests.
__stdin = sys.__stdin__
__stdout = sys.__stdout__
__stderr = sys.__stderr__


class Lightness(enum.Enum):
    """
    Overall color theme of a terminal.

    Can help with deciding which colors to use when printing output.

    """

    UNKNOWN = enum.auto()
    """
    We couldn't determine terminal background, or it wasn't dark
    or bright enough to fall in one category or another.

    """

    DARK = enum.auto()
    """
    Terminal background is dark.

    """

    LIGHT = enum.auto()
    """
    Terminal background is light.

    """


@dataclass(frozen=True, slots=True)
class TerminalTheme:
    """
    Colors and theme of the attached terminal.

    """

    background: yuio.color.ColorValue
    """
    Background color of a terminal.

    """

    foreground: yuio.color.ColorValue
    """
    Foreground color of a terminal.

    """

    black: yuio.color.ColorValue
    """
    Color value for the default "black" color.

    """

    red: yuio.color.ColorValue
    """
    Color value for the default "red" color.

    """

    green: yuio.color.ColorValue
    """
    Color value for the default "green" color.

    """

    yellow: yuio.color.ColorValue
    """
    Color value for the default "yellow" color.

    """

    blue: yuio.color.ColorValue
    """
    Color value for the default "blue" color.

    """

    magenta: yuio.color.ColorValue
    """
    Color value for the default "magenta" color.

    """

    cyan: yuio.color.ColorValue
    """
    Color value for the default "cyan" color.

    """

    white: yuio.color.ColorValue
    """
    Color value for the default "white" color.

    """

    lightness: Lightness
    """
    Overall color theme of a terminal, i.e. dark or light.

    """


@dataclass(frozen=True, slots=True)
class Term:
    """
    This class contains all info about what kinds of things the terminal
    supports. If available, it will also have info about terminal's theme,
    i.e. dark or light background, etc.

    """

    ostream: _t.TextIO
    """
    Terminal's output stream.

    """

    istream: _t.TextIO
    """
    Terminal's input stream.

    """

    color_support: ColorSupport = dataclasses.field(
        default=ColorSupport.NONE, kw_only=True
    )
    """
    Terminal's capability for coloring output.

    """

    ostream_is_tty: bool = dataclasses.field(default=False, kw_only=True)
    """
    Output is connecter to a terminal, and we're not in CI.

    Note that output being connected to a TTY doesn't mean that it's interactive:
    this process can be in background.

    """

    istream_is_tty: bool = dataclasses.field(default=False, kw_only=True)
    """
    Output is connecter to a terminal, and we're not in CI.

    Note that output being connected to a TTY doesn't mean that it's interactive:
    this process can be in background.

    """

    terminal_theme: TerminalTheme | None = dataclasses.field(default=None, kw_only=True)
    """
    Terminal's default foreground, background, and text colors.

    """

    is_unicode: bool = dataclasses.field(default=False, kw_only=True)
    """
    Terminal's output supports unicode characters.

    """

    @property
    def supports_colors(self) -> bool:
        """
        Return :data:`True` if terminal supports simple 8-bit color codes.

        """

        return self.color_support >= ColorSupport.ANSI

    @property
    def supports_colors_256(self) -> bool:
        """
        Return :data:`True` if terminal supports 256-encoded colors.

        """

        return self.color_support >= ColorSupport.ANSI_256

    @property
    def supports_colors_true(self) -> bool:
        """
        Return :data:`True` if terminal supports true colors.

        """

        return self.color_support >= ColorSupport.ANSI_TRUE

    @property
    def is_tty(self) -> bool:
        """
        Return :data:`True` if input and output are connected to a TTY. In this mode
        we can interact with the user by writing and reading lines of text.

        """

        return self.istream_is_tty and self.ostream_is_tty

    @property
    def can_run_widgets(self) -> bool:
        """
        Return :data:`True` if input and output are interactive and colors
        are supported. In this mode we can run interactive widgets.

        """

        return self.color_support >= ColorSupport.ANSI and self.is_tty


_CI_ENV_VARS = [
    "TRAVIS",
    "CIRCLECI",
    "APPVEYOR",
    "GITLAB_CI",
    "BUILDKITE",
    "DRONE",
    "TEAMCITY_VERSION",
    "GITHUB_ACTIONS",
]


def stream_is_unicode(stream: _t.TextIO, /) -> bool:
    """
    Determine of stream's encoding is some version of unicode.

    """

    encoding = getattr(stream, "encoding", None) or locale.getpreferredencoding() or ""
    encoding = encoding.casefold()
    return "utf" in encoding or "unicode" in encoding


def get_tty_size(fallback: tuple[int, int] = (80, 24)):
    """
    Like :func:`shutil.get_terminal_size`, but uses TTY stream if it's available.

    :param fallback:
        tuple with width and height that will be used if query fails.

    """

    _prepare_tty()

    try:
        columns = int(os.environ["COLUMNS"])
    except (KeyError, ValueError):
        columns = 0

    try:
        lines = int(os.environ["LINES"])
    except (KeyError, ValueError):
        lines = 0

    if columns <= 0 or lines <= 0:
        try:
            size = os.get_terminal_size(_TTY_OUTPUT.fileno())  # type: ignore
        except (AttributeError, ValueError, OSError):
            # stream is closed, detached, or not a terminal, or
            # os.get_tty_size() is unsupported
            size = os.terminal_size(fallback)
        if columns <= 0:
            columns = size.columns or fallback[0]
        if lines <= 0:
            lines = size.lines or fallback[1]

    return os.terminal_size((columns, lines))


def detect_ci() -> bool:
    """
    Scan environment variables to detect if we're in a known CI environment.

    """

    return "CI" in os.environ or any(ci in os.environ for ci in _CI_ENV_VARS)


def detect_ci_color_support() -> ColorSupport:
    """
    Scan environment variables to detect an appropriate level of color support
    of a CI environment.

    If we're not in CI, return :attr:`ColorSupport.NONE <yuio.color.ColorSupport.NONE>`.

    """

    if "GITHUB_ACTIONS" in os.environ:
        return ColorSupport.ANSI_TRUE
    elif any(ci in os.environ for ci in _CI_ENV_VARS):
        return ColorSupport.ANSI
    else:
        return ColorSupport.NONE


def get_tty() -> Term:
    """
    Query info about TTY.

    On Unix, this returns terminal connected to ``/dev/tty``. On Windows, this returns
    terminal connected to ``CONIN$``/``CONOUT$``.

    If opening any of these fails, returns :class:`Term` with ``stdin``/``stdout``
    as a fallback.

    .. note::

        Prefer using ``stderr`` for normal IO: your users expect to be able to redirect
        messages from your program.

        Only use ``/dev/tty`` for querying passwords or other things that must not
        be redirected.

    """

    _prepare_tty()
    ostream = _TTY_OUTPUT or __stderr
    istream = _TTY_INPUT or __stdin
    assert ostream is not None
    assert istream is not None
    return get_term_from_stream(ostream, istream)


def get_term_from_stream(
    ostream: _t.TextIO,
    istream: _t.TextIO,
    /,
) -> Term:
    """
    Query info about a terminal attached to the given stream.

    :param ostream:
        output stream.
    :param istream:
        input stream.

    """

    is_unicode = stream_is_unicode(ostream)

    if (
        # For building docs in github.
        "YUIO_FORCE_FULL_TERM_SUPPORT" in os.environ
    ):  # pragma: no cover
        return Term(
            ostream=ostream,
            istream=istream,
            color_support=ColorSupport.ANSI_TRUE,
            ostream_is_tty=True,
            istream_is_tty=True,
            is_unicode=is_unicode,
        )

    _prepare_tty()

    output_is_tty = _output_is_tty(ostream)
    input_is_tty = _input_is_tty(istream)
    in_ci = detect_ci()

    # Detect colors.
    if output_is_tty or _EXPLICIT_COLOR_SUPPORT is not None:
        color_support = _COLOR_SUPPORT
    else:
        color_support = ColorSupport.NONE

    return Term(
        ostream=ostream,
        istream=istream,
        color_support=color_support,
        ostream_is_tty=output_is_tty and not in_ci,
        istream_is_tty=input_is_tty and not in_ci,
        terminal_theme=_TERMINAL_THEME,
        is_unicode=is_unicode,
    )


def _prepare_tty():
    if not _TTY_SETUP_PERFORMED:
        with _LOCK:
            if not _TTY_SETUP_PERFORMED:
                _do_prepare_tty()


def _do_prepare_tty():
    global \
        _TTY_SETUP_PERFORMED, \
        _TERMINAL_THEME, \
        _EXPLICIT_COLOR_SUPPORT, \
        _COLOR_SUPPORT

    _find_tty()

    _TTY_SETUP_PERFORMED = True

    # Theme is `None` for now, will query it later.
    _TERMINAL_THEME = None

    # Find out if user specified `--color` or `FORCE_COLOR`.
    _EXPLICIT_COLOR_SUPPORT = _detect_explicit_color_settings()

    # Check user preferences.
    if _EXPLICIT_COLOR_SUPPORT is False:
        # Colors disabled, nothing more to do.
        _COLOR_SUPPORT = ColorSupport.NONE
        return
    elif _EXPLICIT_COLOR_SUPPORT is True:
        # At least ANSI. Might improve later.
        _COLOR_SUPPORT = max(ColorSupport.ANSI, _detect_color_support_from_env())
    elif _EXPLICIT_COLOR_SUPPORT is None:
        # At least NONE. Might improve later.
        _COLOR_SUPPORT = _detect_color_support_from_env()
    else:
        # Exact color support is given.
        _COLOR_SUPPORT = _EXPLICIT_COLOR_SUPPORT

    if _TTY_OUTPUT is None:
        # Can't find attached TTY output, hence can't improve color support.
        return

    if os.name == "nt":
        # Try enabling true colors.
        if _enable_vt_processing(_TTY_OUTPUT):
            # Success, can improve color support.
            if _EXPLICIT_COLOR_SUPPORT is None or _EXPLICIT_COLOR_SUPPORT is True:
                _COLOR_SUPPORT = ColorSupport.ANSI_TRUE
        else:
            # Failure, this version of Windows does not support colors.
            return

    if _TTY_INPUT is None:
        # Can't find attached TTY input, hence can't improve color support.
        return

    if not _is_foreground(_TTY_OUTPUT) or not _is_foreground(_TTY_INPUT):
        # We're not a foreground process, we won't be able to fetch colors.
        # We also can't query colors if stdin is redirected because
        return
    if detect_ci():
        # We're in CI, we won't be able to fetch colors.
        return
    if not _is_tty(__stdin):
        # We don't want to query colors if our stdin is redirected: this is a sign
        # that this program runs in some sort of a pipeline, and multiple instances
        # of it might run at the same time. If this happens, several processes/threads
        # can interact with the same TTY, leading to garbled output.
        return

    if _COLOR_SUPPORT >= ColorSupport.ANSI:
        # We were able to find TTY, and colors are supported.
        # Try fetching terminal theme.
        _TERMINAL_THEME = _get_standard_colors(_TTY_OUTPUT, _TTY_INPUT)


def _find_tty():
    global _TTY_OUTPUT, _TTY_INPUT

    _TTY_OUTPUT = _TTY_INPUT = None

    closer = contextlib.ExitStack()
    try:
        if os.name == "nt":
            file_io_in = io._WindowsConsoleIO("CONIN$", "r")  # type: ignore
            tty_in = closer.enter_context(
                io.TextIOWrapper(file_io_in, encoding="utf-8")
            )
            file_io_out = io._WindowsConsoleIO("CONOUT$", "w")  # type: ignore
            tty_out = closer.enter_context(
                io.TextIOWrapper(file_io_out, encoding="utf-8")
            )
        else:
            fd = os.open("/dev/tty", os.O_RDWR | os.O_NOCTTY)
            file_io = io.FileIO(fd, "w+")
            file_io.name = "/dev/tty"
            tty_in = tty_out = closer.enter_context(io.TextIOWrapper(file_io))
    except (OSError, AttributeError):
        closer.close()
    except:
        closer.close()
        raise
    else:
        atexit.register(closer.close)
        _TTY_INPUT = tty_in
        _TTY_OUTPUT = tty_out
        return

    for stream in (__stderr, __stdout):
        if stream is not None and _output_is_tty(stream):
            _TTY_OUTPUT = stream
            break
    if __stdin is not None and _input_is_tty(__stdin):
        _TTY_INPUT = __stdin


def _get_standard_colors(
    ostream: _t.TextIO, istream: _t.TextIO
) -> TerminalTheme | None:
    if "YUIO_DISABLE_OSC_QUERIES" in os.environ:
        return None

    try:
        query = "\x1b]10;?\x1b\\\x1b]11;?\x1b\\" + "".join(
            [f"\x1b]4;{i};?\x1b\\" for i in range(8)]
        )
        response = _query_term(ostream, istream, query)
        if not response:
            return None

        # Deal with foreground color.

        match = re.match(
            r"^\x1b]10;rgb:([0-9a-f]{2,4})/([0-9a-f]{2,4})/([0-9a-f]{2,4})(?:\x1b\\|\a)",
            response,
            re.IGNORECASE,
        )
        if match is None:  # pragma: no cover
            return None

        r, g, b = (int(v, 16) // 16 ** (len(v) - 2) for v in match.groups())
        foreground = yuio.color.ColorValue.from_rgb(r, g, b)

        response = response[match.end() :]

        # Deal with background color.

        match = re.match(
            r"^\x1b]11;rgb:([0-9a-f]{2,4})/([0-9a-f]{2,4})/([0-9a-f]{2,4})(?:\x1b\\|\a)",
            response,
            re.IGNORECASE,
        )
        if match is None:  # pragma: no cover
            return None

        r, g, b = (int(v, 16) // 16 ** (len(v) - 2) for v in match.groups())
        background = yuio.color.ColorValue.from_rgb(r, g, b)
        luma = (0.2627 * r + 0.6780 * g + 0.0593 * b) / 256

        if luma <= 0.2:
            lightness = Lightness.DARK
        elif luma >= 0.85:
            lightness = Lightness.LIGHT
        else:
            lightness = Lightness.UNKNOWN

        response = response[match.end() :]

        # Deal with other colors

        colors = {}

        while response:
            match = re.match(
                r"^\x1b]4;(\d+);rgb:([0-9a-f]{2,4})/([0-9a-f]{2,4})/([0-9a-f]{2,4})(?:\x1b\\|\a)",
                response,
                re.IGNORECASE,
            )
            if match is None:  # pragma: no cover
                return None

            c = int(match.group(1))
            r, g, b = (int(v, 16) // 16 ** (len(v) - 2) for v in match.groups()[1:])
            colors[c] = yuio.color.ColorValue.from_rgb(r, g, b)

            response = response[match.end() :]

        if set(colors.keys()) != {0, 1, 2, 3, 4, 5, 6, 7}:  # pragma: no cover
            return None

        # return colors
        return TerminalTheme(
            background=background,
            foreground=foreground,
            black=colors[0],
            red=colors[1],
            green=colors[2],
            yellow=colors[3],
            blue=colors[4],
            magenta=colors[5],
            cyan=colors[6],
            white=colors[7],
            lightness=lightness,
        )

    except Exception:  # pragma: no cover
        return None


def _query_term(ostream: _t.TextIO, istream: _t.TextIO, query: str) -> str | None:
    try:
        with _enter_raw_mode(ostream, istream):
            # Lock the keyboard.
            ostream.write("\x1b[2h")
            ostream.flush()
            _flush_input_buffer(ostream, istream)

            # It is important that we unlock keyboard before exiting `cbreak`,
            # hence the nested `try`.
            try:
                # Append a DA1 query, as virtually all terminals support it.
                ostream.write(query + "\x1b[c")
                ostream.flush()

                buf = _read_keycode(ostream, istream, timeout=0.250)
                if not buf.startswith("\x1b"):
                    yuio._logger.warning("_query_term invalid response: %r", buf)
                    return None

                # Read till we find a DA1 response.
                while not re.search(r"\x1b\[\?.*?c", buf):
                    buf += _read_keycode(ostream, istream, timeout=0.250)

                return buf[: buf.index("\x1b[?")]
            finally:
                _flush_input_buffer(ostream, istream)

                # Release the keyboard.
                ostream.write("\x1b[2i")
                ostream.flush()
    except Exception:  # pragma: no cover
        yuio._logger.warning("_query_term error", exc_info=True)
        return None


def _detect_explicit_color_settings() -> ColorSupport | bool | None:
    color_support = None

    if "FORCE_COLOR" in os.environ:
        color_support = True

    if "NO_COLOR" in os.environ or "FORCE_NO_COLOR" in os.environ:
        color_support = False

    # Note: we don't rely on argparse to parse flags and send them to us
    # because these functions can be called before parsing arguments.
    for arg in sys.argv[1:]:
        if arg in ("--color", "--force-color"):
            color_support = True
        elif arg in ("--no-color", "--force-no-color"):
            color_support = False
        elif arg.startswith(("--color=", "--colors=")):
            value = (
                arg.split("=", maxsplit=1)[1]
                .replace("_", "")
                .replace("-", "")
                .casefold()
            )
            if value in ["1", "yes", "true"]:
                color_support = True
            elif value in ["0", "no", "false"]:
                color_support = False
            elif value == "ansi":
                color_support = ColorSupport.ANSI
            elif value == "ansi256":
                color_support = ColorSupport.ANSI_256
            elif value == "ansitrue":
                color_support = ColorSupport.ANSI_TRUE

    return color_support


def _detect_color_support_from_env() -> ColorSupport:
    term = os.environ.get("TERM", "").lower()
    colorterm = os.environ.get("COLORTERM", "").lower()

    if detect_ci():
        return detect_ci_color_support()
    elif os.name == "nt":
        return ColorSupport.NONE
    elif colorterm in ("truecolor", "24bit") or term == "xterm-kitty":
        return ColorSupport.ANSI_TRUE
    elif colorterm in ("yes", "true") or "256color" in term or term == "screen":
        if os.name == "posix" and term == "xterm-256color" and shutil.which("wslinfo"):
            return ColorSupport.ANSI_TRUE
        else:
            return ColorSupport.ANSI_256
    elif "linux" in term or "color" in term or "ansi" in term or "xterm" in term:
        return ColorSupport.ANSI

    return ColorSupport.NONE


def _is_tty(stream: _t.TextIO | None) -> bool:
    try:
        return stream is not None and stream.isatty()
    except Exception:  # pragma: no cover
        return False


def _input_is_tty(stream: _t.TextIO | None) -> bool:
    try:
        return stream is not None and _is_tty(stream) and stream.readable()
    except Exception:  # pragma: no cover
        return False


def _output_is_tty(stream: _t.TextIO | None) -> bool:
    try:
        return stream is not None and _is_tty(stream) and stream.writable()
    except Exception:  # pragma: no cover
        return False


@contextlib.contextmanager
def _modify_keyboard(
    ostream: _t.TextIO,
    bracketed_paste: bool = False,
    modify_keyboard: bool = False,
):
    prologue = []
    if bracketed_paste:
        prologue.append("\x1b[?2004h")
    if modify_keyboard:
        prologue.append("\x1b[>1u")
    if prologue:
        ostream.write("".join(prologue))
        ostream.flush()
    try:
        yield
    finally:
        epilog = []
        if bracketed_paste:
            epilog.append("\x1b[?2004l")
        epilog.append("\x1b[<u")
        if epilog:
            ostream.write("".join(epilog))
            ostream.flush()


# Platform-specific code for working with terminals.
if os.name == "posix":
    import select
    import signal
    import termios
    import tty

    def _is_foreground(stream: _t.TextIO | None) -> bool:
        try:
            return stream is not None and os.getpgrp() == os.tcgetpgrp(stream.fileno())
        except Exception:  # pragma: no cover
            return False

    @contextlib.contextmanager
    def _enter_raw_mode(
        ostream: _t.TextIO,
        istream: _t.TextIO,
        bracketed_paste: bool = False,
        modify_keyboard: bool = False,
    ):
        prev_mode = termios.tcgetattr(istream)
        new_mode = prev_mode.copy()
        new_mode[tty.LFLAG] &= ~(
            termios.ECHO  # Don't print back what user types.
            | termios.ICANON  # Disable line editing.
            | termios.ISIG  # Disable signals on C-c and C-z.
        )
        new_mode[tty.CC] = new_mode[tty.CC].copy()
        new_mode[tty.CC][termios.VMIN] = 1
        new_mode[tty.CC][termios.VTIME] = 0
        termios.tcsetattr(istream, termios.TCSAFLUSH, new_mode)

        try:
            with _modify_keyboard(ostream, bracketed_paste, modify_keyboard):
                yield
        finally:
            termios.tcsetattr(istream, termios.TCSAFLUSH, prev_mode)

    def _read_keycode(
        ostream: _t.TextIO, istream: _t.TextIO, timeout: float = 0
    ) -> str:
        if timeout and not bool(select.select([istream], [], [], timeout)[0]):
            raise TimeoutError()
        key = os.read(istream.fileno(), 128)
        while bool(select.select([istream], [], [], 0)[0]):
            key += os.read(istream.fileno(), 128)

        return key.decode(istream.encoding, errors="replace")

    def _flush_input_buffer(ostream: _t.TextIO, istream: _t.TextIO):
        pass

    def _enable_vt_processing(ostream: _t.TextIO) -> bool:
        return False  # This is a windows functionality

    def _pause():
        os.kill(os.getpid(), signal.SIGTSTP)

elif os.name == "nt":
    import ctypes
    import ctypes.wintypes
    import msvcrt

    _FlushConsoleInputBuffer = ctypes.windll.kernel32.FlushConsoleInputBuffer
    _FlushConsoleInputBuffer.argtypes = [ctypes.wintypes.HANDLE]
    _FlushConsoleInputBuffer.restype = ctypes.wintypes.BOOL

    _GetConsoleMode = ctypes.windll.kernel32.GetConsoleMode
    _GetConsoleMode.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.LPDWORD]
    _GetConsoleMode.restype = ctypes.wintypes.BOOL

    _SetConsoleMode = ctypes.windll.kernel32.SetConsoleMode
    _SetConsoleMode.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.DWORD]
    _SetConsoleMode.restype = ctypes.wintypes.BOOL

    _GetConsoleWindow = ctypes.windll.kernel32.GetConsoleWindow
    _GetConsoleWindow.argtypes = []
    _GetConsoleWindow.restype = ctypes.wintypes.HWND

    _IsWindowVisible = ctypes.windll.user32.IsWindowVisible
    _IsWindowVisible.argtypes = [ctypes.wintypes.HWND]
    _SetConsoleMode.restype = ctypes.wintypes.BOOL

    _ReadConsoleW = ctypes.windll.kernel32.ReadConsoleW
    _ReadConsoleW.argtypes = [
        ctypes.wintypes.HANDLE,
        ctypes.wintypes.LPVOID,
        ctypes.wintypes.DWORD,
        ctypes.wintypes.LPDWORD,
        ctypes.wintypes.LPVOID,
    ]
    _ReadConsoleW.restype = ctypes.wintypes.BOOL

    _ENABLE_PROCESSED_OUTPUT = 0x0001
    _ENABLE_WRAP_AT_EOL_OUTPUT = 0x0002
    _ENABLE_VIRTUAL_TERMINAL_PROCESSING = 0x0004
    _ENABLE_VIRTUAL_TERMINAL_INPUT = 0x0200

    _ISTREAM_HANDLE = None

    def _is_foreground(stream: _t.TextIO | None) -> bool:
        window = _GetConsoleWindow()
        if not window:
            return False
        return _IsWindowVisible(window)

    @contextlib.contextmanager
    def _enter_raw_mode(
        ostream: _t.TextIO,
        istream: _t.TextIO,
        bracketed_paste: bool = False,
        modify_keyboard: bool = False,
    ):
        global _ISTREAM_HANDLE

        if _ISTREAM_HANDLE is None:
            _prepare_tty()
            _ISTREAM_HANDLE = msvcrt.get_osfhandle((_TTY_INPUT or __stdin).fileno())  # type: ignore

        mode = ctypes.wintypes.DWORD()
        success = _GetConsoleMode(_ISTREAM_HANDLE, ctypes.byref(mode))
        if not success:
            raise ctypes.WinError()
        success = _SetConsoleMode(_ISTREAM_HANDLE, _ENABLE_VIRTUAL_TERMINAL_INPUT)
        if not success:
            raise ctypes.WinError()

        try:
            with _modify_keyboard(ostream, bracketed_paste, modify_keyboard):
                yield
        finally:
            success = _SetConsoleMode(_ISTREAM_HANDLE, mode)
            if not success:
                raise ctypes.WinError()

    def _read_keycode(
        ostream: _t.TextIO, istream: _t.TextIO, timeout: float = 0
    ) -> str:
        assert _ISTREAM_HANDLE is not None

        CHAR16 = ctypes.wintypes.WCHAR * 16

        n_read = ctypes.wintypes.DWORD()
        buffer = CHAR16()

        success = _ReadConsoleW(
            _ISTREAM_HANDLE,
            ctypes.byref(buffer),
            16,
            ctypes.byref(n_read),
            0,
        )
        if not success:
            raise ctypes.WinError()

        return buffer.value

    def _flush_input_buffer(ostream: _t.TextIO, istream: _t.TextIO):
        assert _ISTREAM_HANDLE is not None

        success = _FlushConsoleInputBuffer(_ISTREAM_HANDLE)
        if not success:
            raise ctypes.WinError()

    def _enable_vt_processing(ostream: _t.TextIO) -> bool:
        try:
            version = sys.getwindowsversion()
            if version.major < 10 or version.build < 14931:
                return False

            handle = msvcrt.get_osfhandle(ostream.fileno())
            return bool(
                _SetConsoleMode(
                    handle,
                    _ENABLE_PROCESSED_OUTPUT
                    | _ENABLE_WRAP_AT_EOL_OUTPUT
                    | _ENABLE_VIRTUAL_TERMINAL_PROCESSING,
                )
            )
        except Exception:  # pragma: no cover
            return False

    def _pause():
        pass

else:  # pragma: no cover

    def _is_foreground(stream: _t.TextIO | None) -> bool:
        return False

    @contextlib.contextmanager
    def _enter_raw_mode(
        ostream: _t.TextIO,
        istream: _t.TextIO,
        bracketed_paste: bool = False,
        modify_keyboard: bool = False,
    ):
        raise OSError("not supported")
        yield

    def _read_keycode(
        ostream: _t.TextIO, istream: _t.TextIO, timeout: float = 0
    ) -> str:
        raise OSError("not supported")

    def _flush_input_buffer(ostream: _t.TextIO, istream: _t.TextIO):
        raise OSError("not supported")

    def _enable_vt_processing(ostream: _t.TextIO) -> bool:
        raise OSError("not supported")

    def _pause():
        raise OSError("not supported")



# Yuio project, MIT license.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

"""
This module provides a base class for configs that can be loaded from
files, environment variables or command line arguments (via :mod:`yuio.app`).

Derive your config from the :class:`Config` base class. Inside of its body,
define config fields using type annotations, just like :mod:`dataclasses`:

.. code-block:: python

    class AppConfig(Config):
        #: trained model to execute
        model: pathlib.Path

        #: input data for the model
        data: pathlib.Path

        #: enable or disable gpu
        use_gpu: bool = True

Then use config's constructors and the :meth:`~Config.update` method
to load it from various sources::

    # Load config from a file.
    config = AppConfig.load_from_json_file("~/.my_app_cfg.json")

    # Update config with values from env.
    config.update(AppConfig.load_from_env())


Config base class
-----------------

.. autoclass:: Config


Advanced field configuration
----------------------------

By default, :class:`Config` infers names for env variables and flags,
appropriate parsers, and other things from field's name, type hint, and comments.

If you need to override them, you can use :func:`yuio.app.field`
and :func:`yuio.app.inline` functions (also available from :mod:`yuio.config`):

.. code-block:: python

    class AppConfig(Config):
        model: pathlib.Path | None = field(
            default=None,
            help="trained model to execute",
        )


Nesting configs
---------------

You can nest configs to achieve modularity:

.. code-block:: python

    class ExecutorConfig(Config):
        #: number of threads to use
        threads: int

        #: enable or disable gpu
        use_gpu: bool = True


    class AppConfig(Config):
        #: executor parameters
        executor: ExecutorConfig

        #: trained model to execute
        model: pathlib.Path

To initialise a nested config, pass either an instance of if
or a dict with its variables to the config's constructor:

.. code-block:: python

    # The following lines are equivalent:
    config = AppConfig(executor=ExecutorConfig(threads=16))
    config = AppConfig(executor={"threads": 16})
    # ...although type checkers will complain about dict =(


Parsing environment variables
-----------------------------

You can load config from environment through :meth:`~Config.load_from_env`.

Names of environment variables are just capitalized field names.
Use the :func:`yuio.app.field` function to override them:

.. code-block:: python

    class KillCmdConfig(Config):
        # Will be loaded from `SIGNAL`.
        signal: int

        # Will be loaded from `PROCESS_ID`.
        pid: int = field(env="PROCESS_ID")

In nested configs, environment variable names are prefixed with name
of a field that contains the nested config:

.. code-block:: python

    class BigConfig(Config):
        # `kill_cmd.signal` will be loaded from `KILL_CMD_SIGNAL`.
        kill_cmd: KillCmdConfig

        # `kill_cmd_2.signal` will be loaded from `KILL_SIGNAL`.
        kill_cmd_2: KillCmdConfig = field(env="KILL")

        # `kill_cmd_3.signal` will be loaded from `SIGNAL`.
        kill_cmd_3: KillCmdConfig = field(env="")

You can also disable loading a field from an environment altogether:

.. code-block:: python

    class KillCmdConfig(Config):
        # Will not be loaded from env.
        pid: int = field(env=yuio.DISABLED)

To prefix all variable names with some string, pass the `prefix` parameter
to the :meth:`~Config.load_from_env` function:

.. code-block:: python

    # config.kill_cmd.field will be loaded
    # from `MY_APP_KILL_CMD_SIGNAL`
    config = BigConfig.load_from_env("MY_APP")


Parsing config files
--------------------

You can load config from structured config files,
such as `json`, `yaml` or `toml`:

.. skip: next

.. code-block:: python

    class ExecutorConfig(Config):
        threads: int
        use_gpu: bool = True


    class AppConfig(Config):
        executor: ExecutorConfig
        model: pathlib.Path


    config = AppConfig.load_from_json_file("~/.my_app_cfg.json")

In this example, contents of the above config would be:

.. code-block:: json

    {
        "executor": {
            "threads": 16,
            "use_gpu": true
        },
        "model": "/path/to/model"
    }

Note that, unlike with environment variables,
there is no way to inline nested configs.


Additional config validation
----------------------------

If you have invariants that can't be captured with type system,
you can override :meth:`~Config.validate_config`. This method will be called
every time you load a config from file, arguments or environment:

.. code-block:: python

    class DocGenConfig(Config):
        categories: list[str] = ["quickstart", "api_reference"]
        category_names: dict[str, str] = {"deep_dive": "Deep Dive"}

        def validate_config(self):
            for category in self.category_names:
                if category not in self.categories:
                    raise yuio.parse.ParsingError(f"unknown category {category}")


Merging configs
---------------

Configs are specially designed to be merge-able. The basic pattern is to create
an empty config instance, then :meth:`~Config.update` it with every config source:

.. skip: next

.. code-block:: python

    config = AppConfig()
    config.update(AppConfig.load_from_json_file("~/.my_app_cfg.json"))
    config.update(AppConfig.load_from_env())
    # ...and so on.

The :meth:`~Config.update` function ignores default values, and only overrides
keys that were actually configured.

If you need a more complex update behavior, you can add a merge function for a field:

.. code-block:: python

    class AppConfig(Config):
        plugins: list[str] = field(
            default=[],
            merge=lambda left, right: [*left, *right],
        )

Here, whenever we :meth:`~Config.update` ``AppConfig``, ``plugins`` from both instances
will be concatenated.

.. warning::

    Merge function shouldn't mutate its arguments.
    It should produce a new value instead.

.. warning::

    Merge function will not be called for default value. It's advisable to keep the
    default value empty, and add the actual default to the initial empty config:

    .. skip: next

    .. code-block:: python

        config = AppConfig(plugins=["markdown", "rst"])
        config.update(...)


Re-imports
----------


.. function:: field
    :no-index:

    Alias of :obj:`yuio.app.field`

.. function:: inline
    :no-index:

    Alias of :obj:`yuio.app.inline`

.. function:: positional
    :no-index:

    Alias of :obj:`yuio.app.positional`

.. function:: bool_option
    :no-index:

    Alias of :obj:`yuio.app.bool_option`

.. function:: count_option
    :no-index:

    Alias of :obj:`yuio.app.count_option`

.. function:: parse_many_option
    :no-index:

    Alias of :obj:`yuio.app.parse_many_option`

.. function:: parse_one_option
    :no-index:

    Alias of :obj:`yuio.app.parse_one_option`

.. function:: store_const_option
    :no-index:

    Alias of :obj:`yuio.app.store_const_option`

.. function:: store_false_option
    :no-index:

    Alias of :obj:`yuio.app.store_false_option`

.. function:: store_true_option
    :no-index:

    Alias of :obj:`yuio.app.store_true_option`

.. type:: HelpGroup
    :no-index:

    Alias of :obj:`yuio.cli.HelpGroup`.

.. type:: MutuallyExclusiveGroup
    :no-index:

    Alias of :obj:`yuio.cli.MutuallyExclusiveGroup`.

.. type:: OptionCtor
    :no-index:

    Alias of :obj:`yuio.app.OptionCtor`.

.. type:: OptionSettings
    :no-index:

    Alias of :obj:`yuio.app.OptionSettings`.

.. data:: MISC_GROUP
    :no-index:

    Alias of :obj:`yuio.cli.MISC_GROUP`.

.. data:: OPTS_GROUP
    :no-index:

    Alias of :obj:`yuio.cli.OPTS_GROUP`.

.. data:: SUBCOMMANDS_GROUP
    :no-index:

    Alias of :obj:`yuio.cli.SUBCOMMANDS_GROUP`.

"""

from __future__ import annotations

import json
import os
import pathlib
import textwrap
import types
from dataclasses import dataclass

import yuio
import yuio.cli
import yuio.complete
import yuio.json_schema
import yuio.parse
import yuio.string
from yuio.cli import (
    MISC_GROUP,
    OPTS_GROUP,
    SUBCOMMANDS_GROUP,
    HelpGroup,
    MutuallyExclusiveGroup,
)
from yuio.util import _find_docs

import yuio._typing_ext as _tx
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import typing_extensions as _t
else:
    from yuio import _typing as _t

__all__ = [
    "MISC_GROUP",
    "OPTS_GROUP",
    "SUBCOMMANDS_GROUP",
    "Config",
    "HelpGroup",
    "MutuallyExclusiveGroup",
    "OptionCtor",
    "OptionSettings",
    "bool_option",
    "count_option",
    "field",
    "inline",
    "parse_many_option",
    "parse_one_option",
    "positional",
    "store_const_option",
    "store_false_option",
    "store_true_option",
]

T = _t.TypeVar("T")
Cfg = _t.TypeVar("Cfg", bound="Config")


@dataclass(frozen=True, slots=True)
class _FieldSettings:
    default: _t.Any
    parser: yuio.parse.Parser[_t.Any] | None = None
    help: str | yuio.Disabled | None = None
    env: str | yuio.Disabled | None = None
    flags: str | list[str] | yuio.Positional | yuio.Disabled | None = None
    completer: yuio.complete.Completer | None | yuio.Missing = yuio.MISSING
    metavar: str | None = None
    required: bool | None = None
    merge: _t.Callable[[_t.Any, _t.Any], _t.Any] | None = None
    mutex_group: MutuallyExclusiveGroup | None = None
    help_group: HelpGroup | None = None
    usage: yuio.Group | bool | None = None
    show_if_inherited: bool | None = None
    option_ctor: _t.Callable[[OptionSettings], _t.Any] | None = None

    def _update_defaults(
        self,
        qualname: str,
        name: str,
        ty_with_extras: _t.Any,
        parsed_help: str | None,
        allow_positionals: bool,
    ) -> _Field:
        ty = ty_with_extras
        while _t.get_origin(ty) is _t.Annotated:
            ty = _t.get_args(ty)[0]
        is_subconfig = isinstance(ty, type) and issubclass(ty, Config)

        help: str | yuio.Disabled
        if self.help is not None:
            help = self.help
        elif parsed_help is not None:
            help = parsed_help
        elif is_subconfig and ty.__doc__:
            help = ty.__doc__
        else:
            help = ""

        env: str | yuio.Disabled
        if self.env is not None:
            env = self.env
        else:
            env = name.upper()
        if env == "" and not is_subconfig:
            raise TypeError(f"{qualname} got an empty env variable name")

        flags: list[str] | yuio.Positional | yuio.Disabled
        if self.flags is yuio.DISABLED:
            flags = self.flags
        elif self.flags is yuio.POSITIONAL:
            if not allow_positionals:
                raise TypeError(
                    f"{qualname}: positional arguments are not allowed in configs"
                )
            if is_subconfig:
                raise TypeError(
                    f"error in {qualname}: nested configs can't be positional"
                )
            flags = self.flags
        elif self.flags is None:
            flags = ["--" + name.replace("_", "-")]
        else:
            if isinstance(self.flags, str):
                flags = self.flags.split() or [""]
            else:
                flags = self.flags

            if is_subconfig:
                if not flags:
                    raise TypeError(
                        f"error in {qualname}: nested configs should have exactly one flag; "
                        "to disable prefixing, pass an empty string as a flag"
                    )
                if len(flags) > 1:
                    raise TypeError(
                        f"error in {qualname}: nested configs can't have multiple flags"
                    )
                if flags[0]:
                    if not flags[0].startswith("--"):
                        raise TypeError(
                            f"error in {qualname}: nested configs can't have a short flag"
                        )
                    try:
                        yuio.cli._check_flag(flags[0])
                    except TypeError as e:
                        raise TypeError(f"error in {qualname}: {e}") from None
            else:
                if not flags:
                    raise TypeError(f"{qualname} should have at least one flag")
                for flag in flags:
                    try:
                        yuio.cli._check_flag(flag)
                    except TypeError as e:
                        raise TypeError(f"error in {qualname}: {e}") from None

        default = self.default
        if is_subconfig and default is not yuio.MISSING:
            raise TypeError(f"error in {qualname}: nested configs can't have defaults")

        parser = self.parser
        if is_subconfig and parser is not None:
            raise TypeError(f"error in {qualname}: nested configs can't have parsers")
        elif not is_subconfig and parser is None:
            try:
                parser = yuio.parse.from_type_hint(ty_with_extras)
            except TypeError as e:
                raise TypeError(
                    f"can't derive parser for {qualname}:\n"
                    + textwrap.indent(str(e), "  ")
                ) from None
        if parser is not None:
            origin = _t.get_origin(ty)
            args = _t.get_args(ty)
            is_optional = (
                default is None or _tx.is_union(origin) and types.NoneType in args
            )
            if is_optional and not yuio.parse._is_optional_parser(parser):
                parser = yuio.parse.Optional(parser)
            completer = self.completer
            metavar = self.metavar
            if not metavar and flags is yuio.POSITIONAL:
                metavar = f"<{name.replace('_', '-')}>"
            if completer is not None or metavar is not None:
                parser = yuio.parse.WithMeta(parser, desc=metavar, completer=completer)

        required = self.required
        if is_subconfig and required:
            raise TypeError(f"error in {qualname}: nested configs can't be required")
        if required is None:
            if is_subconfig:
                required = False
            elif allow_positionals:
                required = default is yuio.MISSING
            else:
                required = False

        merge = self.merge
        if is_subconfig and merge is not None:
            raise TypeError(
                f"error in {qualname}: nested configs can't have merge function"
            )

        mutex_group = self.mutex_group
        if is_subconfig and mutex_group is not None:
            raise TypeError(
                f"error in {qualname}: nested configs can't be a part "
                "of a mutually exclusive group"
            )
        if flags is yuio.POSITIONAL and mutex_group is not None:
            raise TypeError(
                f"error in {qualname}: positional arguments can't appear in mutually exclusive groups"
            )

        usage = self.usage

        show_if_inherited = self.show_if_inherited

        help_group = self.help_group

        option_ctor = self.option_ctor
        if option_ctor is not None and is_subconfig:
            raise TypeError(
                f"error in {qualname}: nested configs can't have option constructors"
            )

        return _Field(
            default,
            parser,
            help,
            env,
            flags,
            is_subconfig,
            ty,
            required,
            merge,
            mutex_group,
            help_group,
            usage,
            show_if_inherited,
            option_ctor,
        )


@dataclass(frozen=True, slots=True)
class _Field:
    default: _t.Any
    parser: yuio.parse.Parser[_t.Any] | None
    help: str | yuio.Disabled
    env: str | yuio.Disabled
    flags: list[str] | yuio.Positional | yuio.Disabled
    is_subconfig: bool
    ty: type
    required: bool
    merge: _t.Callable[[_t.Any, _t.Any], _t.Any] | None
    mutex_group: MutuallyExclusiveGroup | None
    help_group: HelpGroup | None
    usage: yuio.Group | bool | None
    show_if_inherited: bool | None = None
    option_ctor: _t.Callable[[OptionSettings], _t.Any] | None = None


@_t.overload
def field(
    *,
    completer: yuio.complete.Completer | None | yuio.Missing = yuio.MISSING,
    metavar: str | None = None,
    help: str | yuio.Disabled | None = None,
    env: str | yuio.Disabled | None = None,
    flags: str | list[str] | yuio.Positional | yuio.Disabled | None = None,
    required: bool | None = None,
    mutex_group: MutuallyExclusiveGroup | None = None,
    help_group: HelpGroup | None = None,
    show_if_inherited: bool | None = None,
    usage: yuio.Group | bool | None = None,
) -> _t.Any: ...
@_t.overload
def field(
    *,
    default: None,
    parser: yuio.parse.Parser[T] | None = None,
    help: str | yuio.Disabled | None = None,
    env: str | yuio.Disabled | None = None,
    flags: str | list[str] | yuio.Positional | yuio.Disabled | None = None,
    required: bool | None = None,
    completer: yuio.complete.Completer | None | yuio.Missing = yuio.MISSING,
    metavar: str | None = None,
    merge: _t.Callable[[T, T], T] | None = None,
    mutex_group: MutuallyExclusiveGroup | None = None,
    help_group: HelpGroup | None = None,
    usage: yuio.Group | bool | None = None,
    show_if_inherited: bool | None = None,
    option_ctor: OptionCtor[T] | None = None,
) -> T | None: ...
@_t.overload
def field(
    *,
    default: T | yuio.Missing = yuio.MISSING,
    parser: yuio.parse.Parser[T] | None = None,
    help: str | yuio.Disabled | None = None,
    env: str | yuio.Disabled | None = None,
    flags: str | list[str] | yuio.Positional | yuio.Disabled | None = None,
    required: bool | None = None,
    completer: yuio.complete.Completer | None | yuio.Missing = yuio.MISSING,
    metavar: str | None = None,
    merge: _t.Callable[[T, T], T] | None = None,
    mutex_group: MutuallyExclusiveGroup | None = None,
    help_group: HelpGroup | None = None,
    usage: yuio.Group | bool | None = None,
    show_if_inherited: bool | None = None,
    option_ctor: OptionCtor[T] | None = None,
) -> T: ...
def field(
    *,
    default: _t.Any = yuio.MISSING,
    parser: yuio.parse.Parser[_t.Any] | None = None,
    help: str | yuio.Disabled | None = None,
    env: str | yuio.Disabled | None = None,
    flags: str | list[str] | yuio.Positional | yuio.Disabled | None = None,
    required: bool | None = None,
    completer: yuio.complete.Completer | None | yuio.Missing = yuio.MISSING,
    metavar: str | None = None,
    merge: _t.Callable[[_t.Any, _t.Any], _t.Any] | None = None,
    mutex_group: MutuallyExclusiveGroup | None = None,
    help_group: HelpGroup | None = None,
    usage: yuio.Group | bool | None = None,
    show_if_inherited: bool | None = None,
    option_ctor: _t.Callable[..., _t.Any] | None = None,
) -> _t.Any:
    """
    Field descriptor, used for additional configuration of CLI options
    and config fields.

    :param default:
        default value for the field or CLI option.
    :param parser:
        parser that will be used to parse config values and CLI options.
    :param help:
        help message that will be used in CLI option description,
        formatted using Markdown (see :mod:`yuio.md`).

        Pass :data:`yuio.DISABLED` to remove this field from CLI help.

        In sub-config fields, controls grouping of fields; pass an empty string
        to disable grouping.
    :param env:
        specifies name of environment variable that will be used if loading config
        from environment.

        Pass :data:`~yuio.DISABLED` to disable loading this field form environment.

        In sub-config fields, controls prefix for all environment variables within
        this sub-config; pass an empty string to disable prefixing.
    :param flags:
        list of names (or a single name) of CLI flags that will be used for this field.

        In configs, pass :data:`~yuio.DISABLED` to disable loading this field form CLI arguments.

        In apps, pass :data:`~yuio.POSITIONAL` to make this argument positional.

        In sub-config fields, controls prefix for all flags withing this sub-config;
        pass an empty string to disable prefixing.
    :param completer:
        completer that will be used for autocompletion in CLI. Using this option
        is equivalent to overriding `completer` with :class:`yuio.parse.WithMeta`.
    :param metavar:
        value description that will be used for CLI help messages. Using this option
        is equivalent to overriding `desc` with :class:`yuio.parse.WithMeta`.
    :param merge:
        defines how values of this field are merged when configs are updated.
    :param mutex_group:
        defines mutually exclusive group for this field.
    :param help_group:
        overrides group in which this field will be placed when generating CLI help
        message.
    :param usage:
        controls how this field renders in CLI usage section.

        Pass :data:`False` to remove this field from usage.

        Pass :class:`yuio.GROUP` to omit this field and add a single string
        ``<options>`` instead.

        Setting `usage` on sub-config fields overrides default `usage` for all
        fields within this sub-config.
    :param option_ctor:
        this parameter is similar to :mod:`argparse`\\ 's ``action``: it allows
        overriding logic for handling CLI arguments by providing a custom
        :class:`~yuio.cli.Option` implementation.

        `option_ctor` should be a callable which takes a single positional argument
        of type :class:`~yuio.app.OptionSettings`, and returns an instance
        of :class:`yuio.cli.Option`.
    :returns:
        a magic object that will be replaced with field's default value once a new
        config class is created.
    :example:
        In apps:

        .. invisible-code-block: python

            import yuio.app

        .. code-block:: python

            @yuio.app.app
            def main(
                # Will be loaded from `--input`.
                input: pathlib.Path | None = None,
                # Will be loaded from `-o` or `--output`.
                output: pathlib.Path | None = field(default=None, flags="-o --output"),
            ): ...

        In configs:

        .. code-block:: python

            class AppConfig(Config):
                model: pathlib.Path | None = field(
                    default=None,
                    help="trained model to execute",
                )

    """

    return _FieldSettings(
        default=default,
        parser=parser,
        help=help,
        env=env,
        flags=flags,
        completer=completer,
        metavar=metavar,
        required=required,
        merge=merge,
        mutex_group=mutex_group,
        help_group=help_group,
        usage=usage,
        show_if_inherited=show_if_inherited,
        option_ctor=option_ctor,
    )


def inline(
    help: str | yuio.Disabled | None = None,
    usage: yuio.Group | bool | None = None,
    show_if_inherited: bool | None = None,
    help_group: HelpGroup | None = None,
) -> _t.Any:
    """
    A shortcut for inlining nested configs.

    Equivalent to calling :func:`~yuio.app.field` with `env` and `flags`
    set to an empty string.

    """

    return field(
        help=help,
        env="",
        flags="",
        usage=usage,
        show_if_inherited=show_if_inherited,
        help_group=help_group,
    )


@_t.overload
def positional(
    *,
    help: str | yuio.Disabled | None = None,
    env: str | yuio.Disabled | None = None,
    completer: yuio.complete.Completer | None | yuio.Missing = yuio.MISSING,
    metavar: str | None = None,
    usage: yuio.Group | bool | None = None,
) -> _t.Any: ...
@_t.overload
def positional(
    *,
    default: None,
    parser: yuio.parse.Parser[T] | None = None,
    help: str | yuio.Disabled | None = None,
    env: str | yuio.Disabled | None = None,
    completer: yuio.complete.Completer | None | yuio.Missing = yuio.MISSING,
    metavar: str | None = None,
    usage: yuio.Group | bool | None = None,
) -> T | None: ...
@_t.overload
def positional(
    *,
    default: T | yuio.Missing = yuio.MISSING,
    parser: yuio.parse.Parser[T] | None = None,
    help: str | yuio.Disabled | None = None,
    env: str | yuio.Disabled | None = None,
    completer: yuio.complete.Completer | None | yuio.Missing = yuio.MISSING,
    metavar: str | None = None,
    usage: yuio.Group | bool | None = None,
) -> T: ...
def positional(
    *,
    default: _t.Any = yuio.MISSING,
    parser: yuio.parse.Parser[_t.Any] | None = None,
    help: str | yuio.Disabled | None = None,
    env: str | yuio.Disabled | None = None,
    completer: yuio.complete.Completer | None | yuio.Missing = yuio.MISSING,
    metavar: str | None = None,
    usage: yuio.Group | bool | None = None,
) -> _t.Any:
    """
    A shortcut for adding a positional argument.

    Equivalent to calling :func:`field` with `flags` set to :data:`~yuio.POSITIONAL`.

    """

    return field(
        default=default,
        parser=parser,
        help=help,
        env=env,
        flags=yuio.POSITIONAL,
        completer=completer,
        metavar=metavar,
        usage=usage,
    )


@_t.dataclass_transform(
    eq_default=False,
    order_default=False,
    kw_only_default=True,
    frozen_default=False,
    field_specifiers=(field, inline, positional),
)
class Config:
    """
    Base class for configs.

    Pass keyword args to set fields, or pass another config to copy it::

        Config(config1, config2, ..., field1=value1, ...)

    Upon creation, all fields that aren't explicitly initialized
    and don't have defaults are considered missing.
    Accessing them will raise :class:`AttributeError`.

    .. automethod:: update

    .. automethod:: load_from_env

    .. automethod:: load_from_json_file

    .. automethod:: load_from_yaml_file

    .. automethod:: load_from_toml_file

    .. automethod:: load_from_parsed_file

    .. automethod:: to_json_schema

    .. automethod:: to_json_value

    .. automethod:: validate_config

    """

    @classmethod
    def __get_fields(cls) -> dict[str, _Field]:
        if cls.__fields is not None:
            return cls.__fields

        if cls.__allow_positionals:
            docs = {}
        else:
            try:
                docs = _find_docs(cls)
            except Exception:
                yuio._logger.warning(
                    "unable to get documentation for class %s.%s",
                    cls.__module__,
                    cls.__qualname__,
                )
                docs = {}

        fields = {}

        for base in reversed(cls.__mro__):
            if base is not cls and hasattr(base, "_Config__get_fields"):
                fields.update(getattr(base, "_Config__get_fields")())

        try:
            types = _t.get_type_hints(cls, include_extras=True)
        except NameError as e:
            if "<locals>" in cls.__qualname__:
                raise NameError(
                    f"{e}. "
                    f"Note: forward references do not work inside functions "
                    f"(see https://github.com/python/typing/issues/797)"
                ) from None
            raise  # pragma: no cover

        for name, field in cls.__gathered_fields.items():
            if not isinstance(field, _FieldSettings):
                field = _FieldSettings(default=field)

            fields[name] = field._update_defaults(
                f"{cls.__qualname__}.{name}",
                name,
                types[name],
                docs.get(name),
                cls.__allow_positionals,
            )
        cls.__fields = fields

        return fields

    def __init_subclass__(cls, _allow_positionals=None, **kwargs):
        super().__init_subclass__(**kwargs)

        if _allow_positionals is not None:
            cls.__allow_positionals: bool = _allow_positionals
        cls.__fields: dict[str, _Field] | None = None

        cls.__gathered_fields: dict[str, _FieldSettings | _t.Any] = {}
        for name in cls.__annotations__:
            if not name.startswith("_"):
                cls.__gathered_fields[name] = cls.__dict__.get(name, yuio.MISSING)
        for name, value in cls.__dict__.items():
            if isinstance(value, _FieldSettings) and name not in cls.__gathered_fields:
                qualname = f"{cls.__qualname__}.{name}"
                raise TypeError(
                    f"error in {qualname}: field without annotations is not allowed"
                )
        for name, value in cls.__gathered_fields.items():
            if isinstance(value, _FieldSettings):
                value = value.default
            setattr(cls, name, value)

    def __init__(self, *args: _t.Self | dict[str, _t.Any], **kwargs):
        for name, field in self.__get_fields().items():
            if field.is_subconfig:
                setattr(self, name, field.ty())

        for arg in args:
            self.update(arg)

        self.update(kwargs)

    def update(self, other: _t.Self | dict[str, _t.Any], /):
        """
        Update fields in this config with fields from another config.

        This function is similar to :meth:`dict.update`.

        Nested configs are updated recursively.

        :param other:
            data for update.

        """

        if not other:
            return

        if isinstance(other, Config):
            if (
                self.__class__ not in other.__class__.__mro__
                and other.__class__ not in self.__class__.__mro__
            ):
                raise TypeError("updating from an incompatible config")
            ns = other.__dict__
        elif isinstance(other, dict):
            ns = other
            for name in ns:
                if name not in self.__get_fields():
                    raise TypeError(f"unknown field: {name}")
        else:
            raise TypeError("expected a dict or a config class")

        for name, field in self.__get_fields().items():
            if name in ns:
                if field.is_subconfig:
                    getattr(self, name).update(ns[name])
                elif ns[name] is not yuio.MISSING:
                    if field.merge is not None and name in self.__dict__:
                        setattr(self, name, field.merge(getattr(self, name), ns[name]))
                    else:
                        setattr(self, name, ns[name])

    @classmethod
    def load_from_env(cls, prefix: str = "") -> _t.Self:
        """
        Load config from environment variables.

        :param prefix:
            if given, names of all environment variables will be prefixed with
            this string and an underscore.
        :returns:
            a parsed config.
        :raises:
            :class:`~yuio.parse.ParsingError`.

        """

        result = cls.__load_from_env(prefix)
        result.validate_config()
        return result

    @classmethod
    def __load_from_env(cls, prefix: str = "") -> _t.Self:
        fields = {}

        for name, field in cls.__get_fields().items():
            if field.env is yuio.DISABLED:
                continue

            if prefix and field.env:
                env = f"{prefix}_{field.env}"
            else:
                env = f"{prefix}{field.env}"

            if field.is_subconfig:
                fields[name] = field.ty.load_from_env(prefix=env)
            elif env in os.environ:
                assert field.parser is not None
                try:
                    fields[name] = field.parser.parse(os.environ[env])
                except yuio.parse.ParsingError as e:
                    raise yuio.parse.ParsingError(
                        "Can't parse environment variable `%s`:\n%s",
                        env,
                        yuio.string.Indent(e),
                    ) from None

        return cls(**fields)

    @classmethod
    def _build_options(cls):
        return cls.__build_options("", "", None, True, False)

    @classmethod
    def __build_options(
        cls,
        prefix: str,
        dest_prefix: str,
        help_group: yuio.cli.HelpGroup | None,
        usage: yuio.Group | bool,
        show_if_inherited: bool,
    ) -> list[yuio.cli.Option[_t.Any]]:
        options: list[yuio.cli.Option[_t.Any]] = []

        if prefix:
            prefix += "-"

        for name, field in cls.__get_fields().items():
            if field.flags is yuio.DISABLED:
                continue

            dest = dest_prefix + name

            flags: list[str] | yuio.Positional
            if prefix and field.flags is not yuio.POSITIONAL:
                flags = [prefix + flag.lstrip("-") for flag in field.flags]
            else:
                flags = field.flags

            field_usage = field.usage
            if field_usage is None:
                field_usage = usage

            field_show_if_inherited = field.show_if_inherited
            if field_show_if_inherited is None:
                field_show_if_inherited = show_if_inherited

            if field.is_subconfig:
                assert flags is not yuio.POSITIONAL
                assert issubclass(field.ty, Config)
                if field.help_group is None:
                    if field.help is yuio.DISABLED:
                        subgroup = yuio.cli.HelpGroup("", help=yuio.DISABLED)
                    elif field.help:
                        lines = field.help.split("\n\n", 1)
                        title = lines[0].replace("\n", " ").rstrip(".") or name
                        help = textwrap.dedent(lines[1]) if len(lines) > 1 else ""
                        subgroup = yuio.cli.HelpGroup(title=title, help=help)
                    else:
                        subgroup = help_group
                else:
                    subgroup = field.help_group
                options.extend(
                    field.ty.__build_options(
                        flags[0],
                        dest + ".",
                        subgroup,
                        field_usage,
                        field_show_if_inherited,
                    )
                )
                continue

            assert field.parser is not None

            option_ctor = field.option_ctor or _default_option
            option = option_ctor(
                OptionSettings(
                    name=name,
                    parser=field.parser,
                    flags=flags,
                    required=field.required,
                    mutex_group=field.mutex_group,
                    usage=field_usage,
                    help=field.help,
                    help_group=field.help_group or help_group,
                    show_if_inherited=field_show_if_inherited,
                    merge=field.merge,
                    dest=dest,
                    default=field.default,
                    long_flag_prefix=prefix or "--",
                )
            )
            options.append(option)

        return options

    @classmethod
    def load_from_json_file(
        cls,
        path: str | pathlib.Path,
        /,
        *,
        ignore_unknown_fields: bool = False,
        ignore_missing_file: bool = False,
    ) -> _t.Self:
        """
        Load config from a ``.json`` file.

        :param path:
            path of the config file.
        :param ignore_unknown_fields:
            if :data:`True`, this method will ignore fields that aren't listed
            in config class.
        :param ignore_missing_file:
            if :data:`True`, silently ignore a missing file error. This is useful
            when loading a config from a home directory.
        :returns:
            a parsed config.
        :raises:
            :class:`~yuio.parse.ParsingError` if config parsing has failed
            or if config file doesn't exist.

        """

        return cls.__load_from_file(
            path, json.loads, ignore_unknown_fields, ignore_missing_file
        )

    @classmethod
    def load_from_yaml_file(
        cls,
        path: str | pathlib.Path,
        /,
        *,
        ignore_unknown_fields: bool = False,
        ignore_missing_file: bool = False,
    ) -> _t.Self:
        """
        Load config from a ``.yaml`` file.

        This requires `PyYaml <https://pypi.org/project/PyYAML/>`__ package
        to be installed.

        :param path:
            path of the config file.
        :param ignore_unknown_fields:
            if :data:`True`, this method will ignore fields that aren't listed
            in config class.
        :param ignore_missing_file:
            if :data:`True`, silently ignore a missing file error. This is useful
            when loading a config from a home directory.
        :returns:
            a parsed config.
        :raises:
            :class:`~yuio.parse.ParsingError` if config parsing has failed
            or if config file doesn't exist. Can raise :class:`ImportError`
            if ``PyYaml`` is not available.

        """

        try:
            import yaml
        except ImportError:
            raise ImportError("PyYaml is not available")

        return cls.__load_from_file(
            path, yaml.safe_load, ignore_unknown_fields, ignore_missing_file
        )

    @classmethod
    def load_from_toml_file(
        cls,
        path: str | pathlib.Path,
        /,
        *,
        ignore_unknown_fields: bool = False,
        ignore_missing_file: bool = False,
    ) -> _t.Self:
        """
        Load config from a ``.toml`` file.

        This requires
        `tomllib <https://docs.python.org/3/library/tomllib.html>`_ or
        `toml <https://pypi.org/project/toml/>`_ package
        to be installed.

        :param path:
            path of the config file.
        :param ignore_unknown_fields:
            if :data:`True`, this method will ignore fields that aren't listed
            in config class.
        :param ignore_missing_file:
            if :data:`True`, silently ignore a missing file error. This is useful
            when loading a config from a home directory.
        :returns:
            a parsed config.
        :raises:
            :class:`~yuio.parse.ParsingError` if config parsing has failed
            or if config file doesn't exist. Can raise :class:`ImportError`
            if ``toml`` is not available.

        """

        try:
            import toml
        except ImportError:
            try:
                import tomllib as toml
            except ImportError:
                raise ImportError("toml is not available")

        return cls.__load_from_file(
            path, toml.loads, ignore_unknown_fields, ignore_missing_file
        )

    @classmethod
    def __load_from_file(
        cls,
        path: str | pathlib.Path,
        file_parser: _t.Callable[[str], _t.Any],
        ignore_unknown_fields: bool = False,
        ignore_missing_file: bool = False,
    ) -> _t.Self:
        path = pathlib.Path(path)

        if ignore_missing_file and (not path.exists() or not path.is_file()):
            return cls()

        try:
            loaded = file_parser(path.read_text())
        except Exception as e:
            raise yuio.parse.ParsingError(
                "Invalid config <c path>%s</c>:\n%s",
                path,
                yuio.string.Indent(e),
            ) from None

        return cls.load_from_parsed_file(
            loaded, ignore_unknown_fields=ignore_unknown_fields, path=path
        )

    @classmethod
    def load_from_parsed_file(
        cls,
        parsed: dict[str, object],
        /,
        *,
        ignore_unknown_fields: bool = False,
        path: str | pathlib.Path | None = None,
    ) -> _t.Self:
        """
        Load config from parsed config file.

        This method takes a dict with arbitrary values that you'd get from
        parsing type-rich configs such as ``yaml`` or ``json``.

        For example::

            with open("conf.yaml") as file:
                config = Config.load_from_parsed_file(yaml.load(file))

        :param parsed:
            data from parsed file.
        :param ignore_unknown_fields:
            if :data:`True`, this method will ignore fields that aren't listed
            in config class.
        :param path:
            path of the original file, used for error reporting.
        :returns:
            a parsed config.
        :raises:
            :class:`~yuio.parse.ParsingError`.

        """

        try:
            result = cls.__load_from_parsed_file(
                yuio.parse.ConfigParsingContext(parsed), ignore_unknown_fields, ""
            )
            result.validate_config()
            return result
        except yuio.parse.ParsingError as e:
            if path is None:
                raise
            else:
                raise yuio.parse.ParsingError(
                    "Invalid config <c path>%s</c>:\n%s",
                    path,
                    yuio.string.Indent(e),
                ) from None

    @classmethod
    def __load_from_parsed_file(
        cls,
        ctx: yuio.parse.ConfigParsingContext,
        ignore_unknown_fields: bool = False,
        field_prefix: str = "",
    ) -> _t.Self:
        value = ctx.value

        if not isinstance(value, dict):
            raise yuio.parse.ParsingError.type_mismatch(value, dict, ctx=ctx)

        fields = {}

        if not ignore_unknown_fields:
            for name in value:
                if name not in cls.__get_fields() and name != "$schema":
                    raise yuio.parse.ParsingError(
                        "Unknown field `%s`", f"{field_prefix}{name}"
                    )

        for name, field in cls.__get_fields().items():
            if name in value:
                if field.is_subconfig:
                    fields[name] = field.ty.__load_from_parsed_file(
                        ctx.descend(value[name], name),
                        ignore_unknown_fields,
                        field_prefix=name + ".",
                    )
                else:
                    assert field.parser is not None
                    fields[name] = field.parser.parse_config_with_ctx(
                        ctx.descend(value[name], name)
                    )

        return cls(**fields)

    def __getattribute(self, item):
        value = super().__getattribute__(item)
        if value is yuio.MISSING:
            raise AttributeError(f"{item} is not configured")
        else:
            return value

    # A dirty hack to hide `__getattribute__` from type checkers.
    locals()["__getattribute__"] = __getattribute

    def __repr__(self):
        field_reprs = ", ".join(
            f"{name}={getattr(self, name, yuio.MISSING)!r}"
            for name in self.__get_fields()
        )
        return f"{self.__class__.__name__}({field_reprs})"

    def __rich_repr__(self):
        for name in self.__get_fields():
            yield name, getattr(self, name, yuio.MISSING)

    def validate_config(self):
        """
        Perform config validation.

        This function is called every time a config is loaded from CLI arguments,
        file, or environment variables. It should check that config is correct,
        and raise :class:`yuio.parse.ParsingError` if it's not.

        :raises:
            :class:`~yuio.parse.ParsingError`.

        """

    @classmethod
    def to_json_schema(
        cls, ctx: yuio.json_schema.JsonSchemaContext
    ) -> yuio.json_schema.JsonSchemaType:
        """
        Create a JSON schema object based on this config.

        The purpose of this method is to make schemas for use in IDEs, i.e. to provide
        autocompletion or simple error checking. The returned schema is not guaranteed
        to reflect all constraints added to the parser.

        :param ctx:
            context for building a schema.
        :returns:
            a JSON schema that describes structure of this config.

        """

        return ctx.add_type(cls, _tx.type_repr(cls), lambda: cls.__to_json_schema(ctx))

    def to_json_value(
        self, *, include_defaults: bool = True
    ) -> yuio.json_schema.JsonValue:
        """
        Convert this config to a representation suitable for JSON serialization.

        :param include_defaults:
            if :data:`False`, default values will be skipped.
        :returns:
            a config converted to JSON-serializable representation.
        :raises:
            :class:`TypeError` if any of the config fields contain values that can't
            be converted to JSON by their respective parsers.

        """

        data = {}
        for name, field in self.__get_fields().items():
            if not include_defaults and name not in self.__dict__:
                continue
            if field.is_subconfig:
                value = getattr(self, name).to_json_value(
                    include_defaults=include_defaults
                )
                if value:
                    data[name] = value
            else:
                assert field.parser
                try:
                    value = getattr(self, name)
                except AttributeError:
                    pass
                else:
                    data[name] = field.parser.to_json_value(value)
        return data

    @classmethod
    def __to_json_schema(
        cls, ctx: yuio.json_schema.JsonSchemaContext
    ) -> yuio.json_schema.JsonSchemaType:
        properties: dict[str, yuio.json_schema.JsonSchemaType] = {}
        defaults = {}

        properties["$schema"] = yuio.json_schema.String()

        for name, field in cls.__get_fields().items():
            if field.is_subconfig:
                properties[name] = field.ty.to_json_schema(ctx)
            else:
                assert field.parser
                field_schema = field.parser.to_json_schema(ctx)
                if field.help and field.help is not yuio.DISABLED:
                    field_schema = yuio.json_schema.Meta(
                        field_schema, description=field.help
                    )
                properties[name] = field_schema
                if field.default is not yuio.MISSING:
                    try:
                        defaults[name] = field.parser.to_json_value(field.default)
                    except TypeError:
                        pass

        return yuio.json_schema.Meta(
            yuio.json_schema.Object(properties),
            title=cls.__name__,
            description=cls.__doc__,
            default=defaults,
        )


Config.__init_subclass__(_allow_positionals=False)


@dataclass(eq=False, kw_only=True)
class OptionSettings:
    """
    Settings for creating an :class:`~yuio.cli.Option` derived from field's type
    and configuration.

    """

    name: str | None
    """
    Name of config field or app parameter that caused creation of this option.

    """

    parser: yuio.parse.Parser[_t.Any]
    """
    Parser associated with this option.

    """

    flags: list[str] | yuio.Positional
    """
    See :attr:`yuio.cli.Option.flags`.

    """

    required: bool
    """
    See :attr:`yuio.cli.Option.required`.

    """

    mutex_group: None | MutuallyExclusiveGroup
    """
    See :attr:`yuio.cli.Option.mutex_group`.

    """

    usage: yuio.Group | bool
    """
    See :attr:`yuio.cli.Option.usage`.

    """

    help: str | yuio.Disabled
    """
    See :attr:`yuio.cli.Option.help`.

    """

    help_group: HelpGroup | None
    """
    See :attr:`yuio.cli.Option.help_group`.

    """

    show_if_inherited: bool
    """
    See :attr:`yuio.cli.Option.show_if_inherited`.

    """

    merge: _t.Callable[[_t.Any, _t.Any], _t.Any] | None
    """
    See :attr:`yuio.cli.ValueOption.merge`.

    """

    dest: str
    """
    See :attr:`yuio.cli.ValueOption.dest`. We don't provide any guarantees about `dest`\\ 's
    contents and recommend treating it as an opaque value.

    """

    default: _t.Any | yuio.Missing
    """
    See :attr:`yuio.cli.ValueOption.default`.

    """

    long_flag_prefix: str
    """
    This argument will contain prefix that was added to all :attr:`~OptionSettings.flags`.
    For apps and top level configs if will be ``"--"``, for nested configs it will
    include additional prefixes, for example ``"--nested-"``.

    """


OptionCtor: _t.TypeAlias = _t.Callable[[OptionSettings], yuio.cli.Option[T]]


def _default_option(s: OptionSettings):
    if s.flags is not yuio.POSITIONAL and yuio.parse._is_bool_parser(s.parser):
        return bool_option()(s)
    elif s.parser.supports_parse_many():
        return parse_many_option()(s)
    else:
        return parse_one_option()(s)


def bool_option(*, neg_flags: list[str] | None = None) -> OptionCtor[bool]:
    """
    Factory for :class:`yuio.cli.BoolOption`.

    :param neg_flags:
        additional set of flags that will set option's value to :data:`False`. If not
        given, a negative flag will be created by adding prefix ``no-`` to the first
        long flag of the option.
    :example:
        Boolean flag :flag:`--json` implicitly creates flag :flag:`--no-json`:

        .. code-block:: python
            :emphasize-lines: 5

            @yuio.app.app
            def main(
                json: bool = yuio.app.field(
                    default=False,
                    option_ctor=yuio.app.bool_option(),
                ),
            ): ...

        Boolean flag :flag:`--json` with explicitly provided flag
        :flag:`--disable-json`:

        .. code-block:: python
            :emphasize-lines: 5-7

            @yuio.app.app
            def main(
                json: bool = yuio.app.field(
                    default=False,
                    option_ctor=yuio.app.bool_option(
                        neg_flags=["--disable-json"],
                    ),
                ),
            ): ...

    """

    def ctor(s: OptionSettings, /):
        if s.flags is yuio.POSITIONAL:
            raise TypeError(f"error in {s.name}: BoolOption can't be positional")
        if neg_flags is None:
            _neg_flags = []
            for flag in s.flags:
                if not yuio.cli._is_short(flag) and flag.startswith(s.long_flag_prefix):
                    prefix = s.long_flag_prefix.strip("-")
                    if prefix:
                        prefix += "-"
                    suffix = flag[len(s.long_flag_prefix) :].removeprefix("-")
                    _neg_flags.append(f"--{prefix}no-{suffix}")
                    break
        elif s.long_flag_prefix == "--":
            _neg_flags = neg_flags
        else:
            _neg_flags = []
            for flag in neg_flags:
                _neg_flags.append(s.long_flag_prefix + flag.lstrip("-"))
        return yuio.cli.BoolOption(
            pos_flags=s.flags,
            neg_flags=_neg_flags,
            required=s.required,
            mutex_group=s.mutex_group,
            usage=s.usage,
            help=s.help,
            help_group=s.help_group,
            show_if_inherited=s.show_if_inherited,
            dest=s.dest,
            parser=s.parser,
            merge=s.merge,
            default=s.default,
        )

    return ctor


def parse_one_option() -> OptionCtor[_t.Any]:
    """
    Factory for :class:`yuio.cli.ParseOneOption`.

    This option takes one argument and passes it
    to :meth:`Parser.parse() <yuio.parse.Parser.parse>`.

    :example:
        Forcing a field which can use :func:`parse_many_option`
        to use :func:`parse_one_option` instead.

        .. code-block:: python
            :emphasize-lines: 6

            @yuio.app.app
            def main(
                files: list[str] = yuio.app.field(
                    default=[],
                    parser=yuio.parse.List(yuio.parse.Int(), delimiter=","),
                    option_ctor=yuio.app.parse_one_option(),
                ),
            ): ...

        This will disable multi-argument syntax:

        .. code-block:: console

            $ prog --files a.txt,b.txt  # Ok
            $ prog --files a.txt b.txt  # Error: `--files` takes one argument.

    """

    def ctor(s: OptionSettings, /):
        return yuio.cli.ParseOneOption(
            flags=s.flags,
            required=s.required,
            mutex_group=s.mutex_group,
            usage=s.usage,
            help=s.help,
            help_group=s.help_group,
            show_if_inherited=s.show_if_inherited,
            dest=s.dest,
            parser=s.parser,
            merge=s.merge,
            default=s.default,
        )

    return ctor


def parse_many_option() -> OptionCtor[_t.Any]:
    """
    Factory for :class:`yuio.cli.ParseManyOption`.

    This option takes multiple arguments and passes them
    to :meth:`Parser.parse_many() <yuio.parse.Parser.parse_many>`.

    """

    def ctor(s: OptionSettings, /):
        return yuio.cli.ParseManyOption(
            flags=s.flags,
            required=s.required,
            mutex_group=s.mutex_group,
            usage=s.usage,
            help=s.help,
            help_group=s.help_group,
            show_if_inherited=s.show_if_inherited,
            dest=s.dest,
            parser=s.parser,
            merge=s.merge,
            default=s.default,
        )

    return ctor


def store_const_option(const: T) -> OptionCtor[T]:
    """
    Factory for :class:`yuio.cli.StoreConstOption`.

    This options takes no arguments. When it's encountered amongst CLI arguments,
    it writes `const` to the resulting config.

    """

    def ctor(s: OptionSettings, /):
        if s.flags is yuio.POSITIONAL:
            raise TypeError(f"error in {s.name}: StoreConstOption can't be positional")

        return yuio.cli.StoreConstOption(
            flags=s.flags,
            required=s.required,
            mutex_group=s.mutex_group,
            usage=s.usage,
            help=s.help,
            help_group=s.help_group,
            show_if_inherited=s.show_if_inherited,
            dest=s.dest,
            merge=s.merge,
            default=s.default,
            const=const,
        )

    return ctor


def count_option() -> OptionCtor[int]:
    """
    Factory for :class:`yuio.cli.CountOption`.

    This option counts number of times it's encountered amongst CLI arguments.

    Equivalent to using :func:`store_const_option` with ``const=1``
    and ``merge=lambda a, b: a + b``.

    :example:

    .. code-block:: python

        @yuio.app.app
        def main(
            quiet: int = yuio.app.field(
                default=0,
                flags=["-q", "--quiet"],
                option_ctor=yuio.app.count_option(),
            ),
        ): ...

    .. code-block:: console

        prog -qq  # quiet=2

    """

    def ctor(s: OptionSettings, /):
        if s.flags is yuio.POSITIONAL:
            raise TypeError(f"error in {s.name}: CountOption can't be positional")

        return yuio.cli.CountOption(
            flags=s.flags,
            required=s.required,
            mutex_group=s.mutex_group,
            usage=s.usage,
            help=s.help,
            help_group=s.help_group,
            show_if_inherited=s.show_if_inherited,
            dest=s.dest,
            default=s.default,
        )

    return ctor


def store_true_option() -> OptionCtor[bool]:
    """
    Factory for :class:`yuio.cli.StoreTrueOption`.

    Equivalent to using :func:`store_const_option` with ``const=True``.

    """

    def ctor(s: OptionSettings, /):
        if s.flags is yuio.POSITIONAL:
            raise TypeError(f"error in {s.name}: StoreTrueOption can't be positional")

        return yuio.cli.StoreTrueOption(
            flags=s.flags,
            required=s.required,
            mutex_group=s.mutex_group,
            usage=s.usage,
            help=s.help,
            help_group=s.help_group,
            show_if_inherited=s.show_if_inherited,
            dest=s.dest,
            default=s.default,
        )

    return ctor


def store_false_option() -> OptionCtor[bool]:
    """
    Factory for :class:`yuio.cli.StoreFalseOption`.

    Equivalent to using :func:`store_const_option` with ``const=False``.

    """

    def ctor(s: OptionSettings, /):
        if s.flags is yuio.POSITIONAL:
            raise TypeError(f"error in {s.name}: StoreFalseOption can't be positional")

        return yuio.cli.StoreFalseOption(
            flags=s.flags,
            required=s.required,
            mutex_group=s.mutex_group,
            usage=s.usage,
            help=s.help,
            help_group=s.help_group,
            show_if_inherited=s.show_if_inherited,
            dest=s.dest,
            default=s.default,
        )

    return ctor



import yuio
import yuio.app
import yuio.config
from yuio.app import App, CommandInfo, app, field, inline, positional

import pytest


@pytest.fixture
def width():
    return 120


@pytest.fixture(autouse=True)
def setup_argv(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(
        "sys.argv", ["prog", "app should not use sys.argv if explicit args are given!"]
    )


@pytest.fixture
def results():
    return {}


class TestBasicArgumentParsing:
    def test_empty_args(self, results):
        @app
        def main(name: str = "", quiet: bool = False, count: int = 0):
            results["name"] = name
            results["quiet"] = quiet
            results["count"] = count

        with pytest.raises(SystemExit) as exc_info:
            main.run([])
        assert exc_info.value.code == 0
        assert results["name"] == ""
        assert results["quiet"] is False
        assert results["count"] == 0

    @pytest.mark.parametrize(
        ("args", "expected_name"),
        [
            (["--name", "Alice"], "Alice"),
            (["--name=Bob"], "Bob"),
        ],
    )
    def test_string_flag(self, results, args, expected_name):
        @app
        def main(name: str = ""):
            results["name"] = name

        with pytest.raises(SystemExit) as exc_info:
            main.run(args)
        assert exc_info.value.code == 0
        assert results["name"] == expected_name

    @pytest.mark.parametrize(
        ("args", "expected"),
        [
            (["--quiet"], True),
            (["--no-quiet"], False),
            (["--quiet=true"], True),
            (["--quiet=false"], False),
        ],
    )
    def test_bool_flag(self, results, args, expected):
        @app
        def main(quiet: bool = False):
            results["quiet"] = quiet

        with pytest.raises(SystemExit) as exc_info:
            main.run(args)
        assert exc_info.value.code == 0
        assert results["quiet"] is expected

    def test_int_flag(self, results):
        @app
        def main(count: int = 0):
            results["count"] = count

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--count", "42"])
        assert exc_info.value.code == 0
        assert results["count"] == 42

    def test_multiple_flags(self, results):
        @app
        def main(name: str = "", count: int = 0, quiet: bool = False):
            results["name"] = name
            results["count"] = count
            results["quiet"] = quiet

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--name", "Eve", "--count", "10", "--quiet"])
        assert exc_info.value.code == 0
        assert results["name"] == "Eve"
        assert results["count"] == 10
        assert results["quiet"] is True


class TestPositionalArguments:
    def test_positional_argument(self, results):
        @app
        def main(input_file: str = positional()):
            results["input_file"] = input_file

        with pytest.raises(SystemExit) as exc_info:
            main.run(["input.txt"])
        assert exc_info.value.code == 0
        assert results["input_file"] == "input.txt"

    def test_positional_with_flag(self, results):
        @app
        def main(input_file: str = positional(), output: str = "out.txt"):
            results["input_file"] = input_file
            results["output"] = output

        with pytest.raises(SystemExit) as exc_info:
            main.run(["input.txt", "--output", "result.txt"])
        assert exc_info.value.code == 0
        assert results["input_file"] == "input.txt"
        assert results["output"] == "result.txt"

    def test_flag_before_positional(self, results):
        @app
        def main(input_file: str = positional(), output: str = "out.txt"):
            results["input_file"] = input_file
            results["output"] = output

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--output", "result.txt", "input.txt"])
        assert exc_info.value.code == 0
        assert results["input_file"] == "input.txt"
        assert results["output"] == "result.txt"

    def test_missing_required_positional(self):
        @app
        def main(input_file: str = positional()):
            pass

        with pytest.raises(SystemExit) as exc_info:
            main.run([])
        assert exc_info.value.code == 1


class TestSubcommands:
    def test_subcommand_basic(self, results):
        @app
        def main():
            results["main_called"] = True

        @main.subcommand
        def build():
            results["build_called"] = True

        with pytest.raises(SystemExit) as exc_info:
            main.run(["build"])
        assert exc_info.value.code == 0
        assert results["main_called"] is True
        assert results["build_called"] is True

    def test_subcommand_with_args(self, results):
        @app
        def main():
            results["main_called"] = True

        @main.subcommand
        def build(target: str = "default"):
            results["target"] = target

        with pytest.raises(SystemExit) as exc_info:
            main.run(["build", "--target", "release"])
        assert exc_info.value.code == 0
        assert results["target"] == "release"

    def test_parent_flag_before_subcommand(self, results):
        @app
        def main(quiet: bool = False):
            results["quiet"] = quiet

        @main.subcommand
        def test_cmd():
            results["test_called"] = True

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--quiet", "test-cmd"])
        assert exc_info.value.code == 0
        assert results["quiet"] is True
        assert results["test_called"] is True

    def test_parent_flag_after_subcommand(self, results):
        @app
        def main(quiet: bool = False):
            results["quiet"] = quiet

        @main.subcommand
        def test_cmd():
            results["test_called"] = True

        with pytest.raises(SystemExit) as exc_info:
            main.run(["test-cmd", "--quiet"])
        assert exc_info.value.code == 0
        assert results["quiet"] is True
        assert results["test_called"] is True

    def test_unknown_subcommand_error(self):
        @app
        def main():
            pass

        @main.subcommand
        def sub():
            pass

        with pytest.raises(SystemExit) as exc_info:
            main.run(["unknown"])
        assert exc_info.value.code == 1

    def test_subcommand_required_by_default(self):
        @app
        def main():
            pass

        @main.subcommand
        def sub():
            pass

        with pytest.raises(SystemExit) as exc_info:
            main.run([])
        assert exc_info.value.code == 1

    def test_optional_subcommand(self, results):
        @app
        def main(quiet: bool = False):
            results["quiet"] = quiet
            results["main_called"] = True

        main.subcommand_required = False

        @main.subcommand
        def sub():
            results["sub_called"] = True

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--quiet"])
        assert exc_info.value.code == 0
        assert results["quiet"] is True
        assert results["main_called"] is True
        assert "sub_called" not in results


class TestSubcommandAliases:
    def test_alias_resolves_to_canonical_name(self, results):
        @app
        def main():
            pass

        @main.subcommand(name="build", aliases=["b"])
        def build_cmd():
            results["build_called"] = True

        with pytest.raises(SystemExit) as exc_info:
            main.run(["b"])
        assert exc_info.value.code == 0
        assert results["build_called"] is True

        results.clear()
        with pytest.raises(SystemExit) as exc_info:
            main.run(["build"])
        assert exc_info.value.code == 0
        assert results["build_called"] is True


class TestNestedSubcommands:
    def test_two_level_subcommands(self, results):
        @app
        def main(root_flag: bool = False):
            results["root_flag"] = root_flag

        @main.subcommand
        def outer(outer_flag: str = ""):
            results["outer_flag"] = outer_flag

        @outer.subcommand
        def inner(inner_flag: int = 0):
            results["inner_flag"] = inner_flag

        with pytest.raises(SystemExit) as exc_info:
            main.run(
                [
                    "--root-flag",
                    "outer",
                    "--outer-flag",
                    "val",
                    "inner",
                    "--inner-flag",
                    "42",
                ]
            )
        assert exc_info.value.code == 0
        assert results["root_flag"] is True
        assert results["outer_flag"] == "val"
        assert results["inner_flag"] == 42


class TestCommandInfo:
    def test_command_info_name(self, results):
        @app
        def main(_command_info: CommandInfo):
            results["name"] = _command_info.name

        with pytest.raises(SystemExit) as exc_info:
            main.run([])
        assert exc_info.value.code == 0
        assert results["name"] == "__main__"

    def test_command_info_subcommand_none_when_no_subcommand(self, results):
        @app
        def main(_command_info: CommandInfo):
            results["subcommand"] = _command_info.subcommand

        with pytest.raises(SystemExit) as exc_info:
            main.run([])
        assert exc_info.value.code == 0
        assert results["subcommand"] is None

    def test_command_info_subcommand_present(self, results):
        @app
        def main(_command_info: CommandInfo):
            results["has_subcommand"] = _command_info.subcommand is not None
            if _command_info.subcommand:
                results["subcommand_name"] = _command_info.subcommand.name

        main.subcommand_required = False

        @main.subcommand
        def sub():
            results["sub_called"] = True

        with pytest.raises(SystemExit) as exc_info:
            main.run(["sub"])
        assert exc_info.value.code == 0
        assert results["has_subcommand"] is True
        assert results["subcommand_name"] == "sub"
        assert results["sub_called"] is True

    def test_command_info_manual_subcommand_invocation(self, results):
        @app
        def main(_command_info: CommandInfo):
            results["main_called"] = True
            if _command_info.subcommand:
                _command_info.subcommand()
            return False  # Prevent automatic subcommand invocation

        main.subcommand_required = False

        @main.subcommand
        def sub():
            results["sub_called"] = True
            results["sub_called_n"] = results.setdefault("sub_called_n", 0) + 1

        with pytest.raises(SystemExit) as exc_info:
            main.run(["sub"])
        assert exc_info.value.code == 0
        assert results["main_called"] is True
        assert results["sub_called"] is True
        assert results["sub_called_n"] == 1

    def test_command_info_manual_subcommand_invocation_without_explicit_prevention(
        self, results
    ):
        @app
        def main(_command_info: CommandInfo):
            results["main_called"] = True
            if _command_info.subcommand:
                _command_info.subcommand()
            # Don't prevent automatic subcommand invocation

        main.subcommand_required = False

        @main.subcommand
        def sub():
            results["sub_called"] = True
            results["sub_called_n"] = results.setdefault("sub_called_n", 0) + 1

        with pytest.raises(SystemExit) as exc_info:
            main.run(["sub"])
        assert exc_info.value.code == 0
        assert results["main_called"] is True
        assert results["sub_called"] is True
        assert results["sub_called_n"] == 1

    def test_command_info_prevent_subcommand_invocation(self, results):
        @app
        def main(_command_info: CommandInfo):
            results["main_called"] = True
            return False  # Prevent automatic subcommand invocation

        main.subcommand_required = False

        @main.subcommand
        def sub():
            results["sub_called"] = True

        with pytest.raises(SystemExit) as exc_info:
            main.run(["sub"])
        assert exc_info.value.code == 0
        assert results["main_called"] is True
        assert "sub_called" not in results


class TestFieldFunction:
    def test_field_with_custom_flags(self, results):
        @app
        def main(output: str = field(default="", flags=["-o", "--output"])):
            results["output"] = output

        with pytest.raises(SystemExit) as exc_info:
            main.run(["-o", "test.txt"])
        assert exc_info.value.code == 0
        assert results["output"] == "test.txt"

    def test_field_required(self):
        @app
        def main(name: str = field(required=True)):
            pass

        with pytest.raises(SystemExit) as exc_info:
            main.run([])
        assert exc_info.value.code == 1


class TestInlineFunction:
    def test_inline_config(self, results):
        class DbConfig(yuio.config.Config):
            host: str = "localhost"
            port: int = 5432

        @app
        def main(db: DbConfig = inline()):
            results["host"] = db.host
            results["port"] = db.port

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--host", "example.com", "--port", "3306"])
        assert exc_info.value.code == 0
        assert results["host"] == "example.com"
        assert results["port"] == 3306


class TestConfigIntegration:
    def test_nested_config_with_prefix(self, results):
        class AuthConfig(yuio.config.Config):
            token: str = ""
            timeout: int = 30

        @app
        def main(auth: AuthConfig = field(flags="--auth")):
            results["token"] = auth.token
            results["timeout"] = auth.timeout

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--auth-token", "secret", "--auth-timeout", "60"])
        assert exc_info.value.code == 0
        assert results["token"] == "secret"
        assert results["timeout"] == 60


class TestEdgeCases:
    def test_empty_string_value(self, results):
        @app
        def main(name: str = "default"):
            results["name"] = name

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--name="])
        assert exc_info.value.code == 0
        assert results["name"] == ""

    def test_value_with_equals_sign(self, results):
        @app
        def main(config: str = ""):
            results["config"] = config

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--config=key=value"])
        assert exc_info.value.code == 0
        assert results["config"] == "key=value"

    def test_negative_numbers(self, results):
        @app
        def main(count: int = 0):
            results["count"] = count

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--count", "-5"])
        assert exc_info.value.code == 0
        assert results["count"] == -5

    def test_double_dash_separator(self, results):
        @app
        def main(flag: bool = False, files: list[str] = positional(default=[])):
            results["flag"] = flag
            results["files"] = files

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--", "--flag", "-x"])
        assert exc_info.value.code == 0
        assert results["flag"] is False
        assert results["files"] == ["--flag", "-x"]

    def test_unknown_flag_error(self):
        @app
        def main(name: str = ""):
            pass

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--unknown"])
        assert exc_info.value.code == 1


class TestListArguments:
    def test_list_multiple_values(self, results):
        @app
        def main(files: list[str] = []):
            results["files"] = files

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--files", "a.txt", "b.txt", "c.txt"])
        assert exc_info.value.code == 0
        assert results["files"] == ["a.txt", "b.txt", "c.txt"]

    def test_list_inline_value(self, results):
        @app
        def main(files: list[str] = []):
            results["files"] = files

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--files=a.txt b.txt c.txt"])
        assert exc_info.value.code == 0
        assert results["files"] == ["a.txt", "b.txt", "c.txt"]

    def test_list_empty(self, results):
        @app
        def main(files: list[str] = []):
            results["files"] = files

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--files"])
        assert exc_info.value.code == 0
        assert results["files"] == []

    def test_list_specified_two_times(self, results):
        @app
        def main(files: list[str] = []):
            results["files"] = files

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--files", "a.txt", "b.txt", "--files", "x.txt", "y.txt"])
        assert exc_info.value.code == 0
        assert results["files"] == ["x.txt", "y.txt"]


class TestExitCodes:
    def test_unknown_flag_exits_with_1(self):
        @app
        def main(name: str = ""):
            pass

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--unknown"])
        assert exc_info.value.code == 1

    def test_missing_required_exits_with_1(self):
        @app
        def main(name: str = field(required=True)):
            pass

        with pytest.raises(SystemExit) as exc_info:
            main.run([])
        assert exc_info.value.code == 1

    def test_invalid_value_exits_with_1(self):
        @app
        def main(count: int = 0):
            pass

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--count", "not_a_number"])
        assert exc_info.value.code == 1

    def test_unknown_subcommand_exits_with_1(self):
        @app
        def main():
            pass

        @main.subcommand
        def sub():
            pass

        with pytest.raises(SystemExit) as exc_info:
            main.run(["unknown_sub"])
        assert exc_info.value.code == 1

    def test_app_error_exits_with_1(self):
        @app
        def main():
            raise yuio.app.AppError("Something went wrong")

        with pytest.raises(SystemExit) as exc_info:
            main.run([])
        assert exc_info.value.code == 1

    def test_keyboard_interrupt_exits_with_130(self):
        @app
        def main():
            raise KeyboardInterrupt()

        with pytest.raises(SystemExit) as exc_info:
            main.run([])
        assert exc_info.value.code == 130

    def test_error_exits_with_3(self):
        @app
        def main():
            raise RuntimeError()

        with pytest.raises(SystemExit) as exc_info:
            main.run([])
        assert exc_info.value.code == 3


class TestMutuallyExclusiveGroups:
    def test_mutex_group_error(self):
        group = yuio.config.MutuallyExclusiveGroup()

        @app
        def main(
            json_output: bool = field(default=False, mutex_group=group),
            xml_output: bool = field(default=False, mutex_group=group),
        ):
            pass

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--json-output", "--xml-output"])
        assert exc_info.value.code == 1

    def test_mutex_group_single_allowed(self, results):
        group = yuio.config.MutuallyExclusiveGroup()

        @app
        def main(
            json_output: bool = field(default=False, mutex_group=group),
            xml_output: bool = field(default=False, mutex_group=group),
        ):
            results["json_output"] = json_output
            results["xml_output"] = xml_output

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--json-output"])
        assert exc_info.value.code == 0
        assert results["json_output"] is True
        assert results["xml_output"] is False

    def test_mutex_group_required(self):
        group = yuio.config.MutuallyExclusiveGroup(required=True)

        @app
        def main(
            json_output: bool = field(default=False, mutex_group=group),
            xml_output: bool = field(default=False, mutex_group=group),
        ):
            pass

        with pytest.raises(SystemExit) as exc_info:
            main.run([])
        assert exc_info.value.code == 1


class TestAppSettings:
    def test_app_prog_name(self):
        @app(prog="myapp")
        def main():
            pass

        assert main.prog == "myapp"

    def test_app_description_from_docstring(self):
        @app
        def main():
            """This is the app description."""
            pass

        assert main.description is not None
        assert "app description" in main.description

    def test_app_version(self):
        @app(version="1.0.0")
        def main():
            pass

        assert main.version == "1.0.0"

    def test_app_epilog(self):
        @app(epilog="More info at example.com")
        def main():
            pass

        assert main.epilog == "More info at example.com"


class TestSpecialParameters:
    def test_unknown_special_param_error(self):
        with pytest.raises(TypeError, match="unknown special param"):

            @app
            def main(_unknown_param: str = ""):
                pass

    def test_command_info_param_accepted(self, results):
        @app
        def main(_command_info: CommandInfo | None):
            results["info_received"] = _command_info is not None

        with pytest.raises(SystemExit) as exc_info:
            main.run([])
        assert exc_info.value.code == 0
        assert results["info_received"] is True


class TestOptionCtors:
    def test_count_option(self, results):
        @app
        def main(
            quiet: int = field(
                default=0,
                flags=["-q", "--quiet"],
                option_ctor=yuio.app.count_option(),
            ),
        ):
            results["quiet"] = quiet

        with pytest.raises(SystemExit) as exc_info:
            main.run(["-q", "-q", "-q"])
        assert exc_info.value.code == 0
        assert results["quiet"] == 3

    def test_store_true_option(self, results):
        @app
        def main(
            debug: bool = field(
                default=False,
                option_ctor=yuio.app.store_true_option(),
            ),
        ):
            results["debug"] = debug

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--debug"])
        assert exc_info.value.code == 0
        assert results["debug"] is True

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--debug=true"])
        assert exc_info.value.code == 1

    def test_store_false_option(self, results):
        @app
        def main(
            cache: bool = field(
                default=True,
                flags=["--no-cache"],
                option_ctor=yuio.app.store_false_option(),
            ),
        ):
            results["cache"] = cache

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--no-cache"])
        assert exc_info.value.code == 0
        assert results["cache"] is False

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--no-cache=true"])
        assert exc_info.value.code == 1

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--no-no-cache"])
        assert exc_info.value.code == 1

    def test_store_const_option(self, results):
        @app
        def main(
            format: str = field(
                default="text",
                flags=["--json"],
                option_ctor=yuio.app.store_const_option("json"),
            ),
        ):
            results["format"] = format

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--json"])
        assert exc_info.value.code == 0
        assert results["format"] == "json"

    def test_bool_option(self, results):
        @app
        def main(
            json_output: bool = field(
                default=False,
                option_ctor=yuio.app.bool_option(),
            ),
        ):
            results["json_output"] = json_output

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--json-output"])
        assert exc_info.value.code == 0
        assert results["json_output"] is True

    def test_bool_option_custom_neg_flags(self, results):
        @app
        def main(
            json_output: bool = field(
                default=False,
                flags=["--json"],
                option_ctor=yuio.app.bool_option(neg_flags=["--disable-json"]),
            ),
        ):
            results["json_output"] = json_output

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--disable-json"])
        assert exc_info.value.code == 0
        assert results["json_output"] is False

    def test_bool_option_auto_generated_neg_flag(self, results):
        @app
        def main(
            feature: bool | None = field(
                default=None,
                flags=["--enable-feature"],
                option_ctor=yuio.app.bool_option(),
            ),
        ):
            results["feature"] = feature

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--no-enable-feature"])
        assert exc_info.value.code == 0
        assert results["feature"] is False

    def test_bool_option_auto_generated_neg_flag_with_prefix(self, results):
        class ConnectionConfig(yuio.config.Config):
            keep_alive: bool | None = field(
                default=None,
                option_ctor=yuio.app.bool_option(),
            )

        @app
        def main(
            connection: ConnectionConfig,
        ):
            results["keep_alive"] = connection.keep_alive

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--connection-no-keep-alive"])
        assert exc_info.value.code == 0
        assert results["keep_alive"] is False

    def test_bool_option_positional_raises(self):
        @app
        def main(
            flag: bool = field(
                default=False,
                flags=yuio.POSITIONAL,
                option_ctor=yuio.app.bool_option(),
            ),
        ):
            pass

        with pytest.raises(SystemExit) as exc_info:
            main.run([])
        assert exc_info.value.code == 3


class TestShortFlagCombinations:
    def test_combined_short_flags(self, results):
        @app
        def main(
            all_files: bool = field(default=False, flags=["-a"]),
            long_format: bool = field(default=False, flags=["-l"]),
            human_readable: bool = field(default=False, flags=["-H"]),
        ):
            results["all_files"] = all_files
            results["long_format"] = long_format
            results["human_readable"] = human_readable

        with pytest.raises(SystemExit) as exc_info:
            main.run(["-alH"])
        assert exc_info.value.code == 0
        assert results["all_files"] is True
        assert results["long_format"] is True
        assert results["human_readable"] is True


class TestAppDecorator:
    def test_app_decorator_no_parens(self):
        @app
        def main():
            pass

        assert isinstance(main, App)

    def test_app_decorator_with_parens(self):
        @app()
        def main():
            pass

        assert isinstance(main, App)

    def test_app_decorator_with_args(self):
        @app(prog="test", version="1.0")
        def main():
            pass

        assert isinstance(main, App)
        assert main.prog == "test"
        assert main.version == "1.0"


class TestAppWrapped:
    def test_wrapped_callable(self, results):
        @app
        def main(name: str = "default"):
            results["name"] = name

        result = main.wrapped(name="test")
        assert result is False  # Returns False from CommandInfo()
        assert results["name"] == "test"

    def test_wrapped_with_positional_args(self, results):
        @app
        def main(a: str = "", b: int = 0):
            results["a"] = a
            results["b"] = b

        result = main.wrapped("hello", 42)
        assert result is False
        assert results["a"] == "hello"
        assert results["b"] == 42

    def test_wrapped_with_fields(self, results):
        @app
        def main(name: str = field(default="default")):
            results["name"] = name

        result = main.wrapped(name="test")
        assert result is False  # Returns False from CommandInfo()
        assert results["name"] == "test"

    def test_wrapped_with_command_info(self, results):
        @app
        def main(_command_info, name: str = field(default="default")):
            results["name"] = name

        result = main.wrapped(name="test")  # type: ignore
        assert result is False  # Returns False from CommandInfo()
        assert results["name"] == "test"


class TestHelpAndVersionFlags:
    def test_help_flag_exits_with_0(self, capsys):
        @app
        def main():
            pass

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--help"])
        assert exc_info.value.code == 0

    def test_version_flag_exits_with_0(self, capsys):
        @app(version="1.0.0")
        def main():
            pass

        with pytest.raises(SystemExit) as exc_info:
            main.run(["--version"])
        assert exc_info.value.code == 0



import pathlib
import yuio.app

CONFIG_GROUP = yuio.app.HelpGroup("Config")

@yuio.app.app
def main(
    #: Override default path to config.
    config: pathlib.Path | None = yuio.app.field(
        default=None,
        help_group=CONFIG_GROUP,
    ),
    #: Print output in machine readable format.
    machine_readable: bool = yuio.app.field(
        default=False,
        help_group=yuio.app.MISC_GROUP,
    ),
):
    ...

if __name__ == "__main__":
    main.run()



import yuio.io

if __name__ == "__main__":
    yuio.io.info(
        "Enter some ints separated by spaces. "
        "Try entering invalid ints and see error messages."
    )

    value = yuio.io.ask[list[int]]("Enter numbers")

    yuio.io.success("You've entered %#r.", value)



import logging
import os
import pathlib

import yuio.exec

import pytest


def make_script(exitcode=0, as_shell=False):
    if os.name == "nt":
        command = f"echo out_message & echo err_message 1>&2 & exit {exitcode}"
        if as_shell:
            return [command]
        else:
            return ["cmd", "/k", command]
    else:
        command = f"echo out_message && echo err_message 1>&2 && exit {exitcode}"
        if as_shell:
            return [command]
        else:
            return ["sh", "-c", command]


@pytest.mark.parametrize(
    ("logger", "logger_name"),
    [
        (None, "yuio.exec"),
        ("custom.logger", "custom.logger"),
        (logging.getLogger("custom.logger"), "custom.logger"),
    ],
)
def test_exec(capsys, caplog, logger, logger_name):
    caplog.set_level(logging.DEBUG)

    result = yuio.exec.exec(*make_script(), logger=logger)
    assert result.strip() == "out_message"

    result = yuio.exec.exec(*make_script(), logger=logger, level=logging.INFO)
    assert result.strip() == "out_message"

    res = capsys.readouterr()
    assert res.out == ""
    assert res.err == ""

    records = [(n, l, m.strip()) for n, l, m in caplog.record_tuples]
    assert (logger_name, logging.DEBUG, "-> err_message") in records
    assert (logger_name, logging.INFO, "-> err_message") in records


def test_input():
    if os.name == "nt":
        args = [
            "powershell",
            "-noprofile",
            "-command",
            "[Console]::In.ReadToEnd() | Write-Host -NoNewline",
        ]
    else:
        args = ["cat"]
    result = yuio.exec.exec(*args, input="hello\n")
    assert result.strip() == "hello"


def test_large_input():
    if os.name == "nt":
        args = [
            "powershell",
            "-noprofile",
            "-command",
            "[Console]::In.ReadToEnd() | Write-Host -NoNewline",
        ]
    else:
        args = ["cat"]
    data = "x" * 1024 * 1024
    result = yuio.exec.exec(*args, input=data + "\n")
    assert result.strip() == data


def test_env():
    if os.name == "nt":
        args = ["cmd", "/k", "echo %FOO% & exit 0"]
    else:
        args = ["printenv", "FOO"]

    result = yuio.exec.exec(*args, env={"FOO": "BAR"})
    assert result.strip() == "BAR"


def test_cwd(tmp_path):
    if os.name == "nt":
        args = ["cmd", "/k", "cd & exit 0"]
    else:
        args = ["pwd"]
    result = yuio.exec.exec(*args, cwd=tmp_path)
    assert result.strip() == str(tmp_path)


@pytest.mark.linux
@pytest.mark.darwin
def test_path(tmp_path):
    yuio.exec.exec("ls", pathlib.Path(tmp_path))


def test_fail():
    try:
        yuio.exec.exec(*make_script(2))
    except yuio.exec.ExecError as e:
        assert "err_message" in e.stderr
        assert e.stdout.strip() == "out_message"
        assert e.returncode == 2
    else:
        assert False, "didn't fail"


def test_dont_capture_io(capsys, caplog):
    caplog.set_level(logging.DEBUG, "yuio.exec")

    result = yuio.exec.exec(*make_script(), capture_io=False)
    assert result is None

    # We check that our loggers didn't print anything.
    res = capsys.readouterr()
    assert res.out == ""
    assert res.err == ""
    assert " ".join(make_script()) in caplog.messages



import os
import pathlib
import sys
import time

import pytest
from libtmux.pane import Pane
from libtmux.server import Server


@pytest.fixture
def python_path():
    return pathlib.Path(sys.executable)


@pytest.fixture
def pane(server: Server):
    session = server.new_session()
    for k, v in os.environ.items():
        if k.startswith("COV"):
            session.set_environment(k, v)
    window = session.new_window(window_shell="bash")
    pane = window.active_pane
    assert pane is not None
    return pane


@pytest.mark.linux
@pytest.mark.full
@pytest.mark.flaky
def test_tmux(pane: Pane, tmp_path, python_path):
    output_path = tmp_path / "output.txt"
    pane.send_keys("pwd")
    pane.send_keys(
        f"{python_path} -m yuio.scripts.showkey --modify-keyboard > {output_path}"
    )
    time.sleep(2)
    pane.send_keys("abc")
    keys = [
        "Escape",
        "Up",
        "Down",
        "Left",
        "Right",
        "Home",
        "End",
        "PageUp",
        "PageDown",
        "Insert",
        "Delete",
        "BSpace",
        "Tab",
        "BTab",
        "Space",
    ]
    prefixes = [
        "",
        "^",
        "M-",
        "^M-",
    ]
    for key in keys:
        for prefix in prefixes:
            pane.send_keys(prefix + key, enter=False)
    pane.send_keys("C-d", enter=False)
    pane.send_keys("C-5", enter=False)
    pane.send_keys("C-_", enter=False)
    pane.send_keys("C-c", enter=False)
    print("-" * 80)
    print("\n".join(pane.capture_pane()))
    print("-" * 80)
    pane.kill()
    result = output_path.read_text()
    print(result)
    assert result == EXPECTED


EXPECTED = r"""
Key: 'a'
Key: 'b'
Key: 'c'
Key: ENTER
Key: ESCAPE
Key: Ctrl+ESCAPE
Key: Alt+ESCAPE
Key: Ctrl+Alt+ESCAPE
Key: ARROW_UP
Key: Ctrl+ARROW_UP
Key: Alt+ARROW_UP
Key: Ctrl+Alt+ARROW_UP
Key: ARROW_DOWN
Key: Ctrl+ARROW_DOWN
Key: Alt+ARROW_DOWN
Key: Ctrl+Alt+ARROW_DOWN
Key: ARROW_LEFT
Key: Ctrl+ARROW_LEFT
Key: Alt+ARROW_LEFT
Key: Ctrl+Alt+ARROW_LEFT
Key: ARROW_RIGHT
Key: Ctrl+ARROW_RIGHT
Key: Alt+ARROW_RIGHT
Key: Ctrl+Alt+ARROW_RIGHT
Key: HOME
Key: Ctrl+HOME
Key: Alt+HOME
Key: Ctrl+Alt+HOME
Key: END
Key: Ctrl+END
Key: Alt+END
Key: Ctrl+Alt+END
Key: PAGE_UP
Key: Ctrl+PAGE_UP
Key: Alt+PAGE_UP
Key: Ctrl+Alt+PAGE_UP
Key: PAGE_DOWN
Key: Ctrl+PAGE_DOWN
Key: Alt+PAGE_DOWN
Key: Ctrl+Alt+PAGE_DOWN
Key: INSERT
Key: Ctrl+INSERT
Key: Alt+INSERT
Key: Ctrl+Alt+INSERT
Key: DELETE
Key: Ctrl+DELETE
Key: Alt+DELETE
Key: Ctrl+Alt+DELETE
Key: BACKSPACE
Key: Ctrl+BACKSPACE
Key: Alt+BACKSPACE
Key: Ctrl+Alt+BACKSPACE
Key: TAB
Key: Ctrl+TAB
Key: Alt+TAB
Key: Alt+TAB
Key: Shift+TAB
Key: Ctrl+'\U0010e1a7'
Key: Shift+Alt+TAB
Key: Ctrl+Alt+'\U0010e1a7'
Key: ' '
Key: Ctrl+'`'
Key: Alt+' '
Key: Ctrl+Alt+'`'
Key: Ctrl+'d'
Key: Ctrl+'5'
Key: Ctrl+'7'
""".lstrip()
